from collections import defaultdict
from enum import Enum
from typing import List

from sidermit.city import Graph, Node
from sidermit.publictransportsystem import TransportMode, Route
from sidermit.publictransportsystem.network import RouteType


#                                    Representation extended graph
#
# for each p,sc y CBD:                _______________city_node_______________
#                                     ↑                  ↑                   ↑
#                                access_edge        access_edge          access_edge
#                                     ↓                  ↓                   ↓
# for each mode:                  stop_node (mode1)  stop_node (mode2)    stop_node (mode3)
#                                 |       ↑          |       ↑            |       ↑
#                             boarding alighting  boarding alighting   boarding alighting
#                               edge      edge      edge     edge        edge     edge
#                                 ↓       |          ↓       |            ↓       |
# for each route, direction      route_node           route_node          route_node
#                                     ↑                    ↑                   ↑
#                                route_edge             route_edge         route_edge
#                                     ↑                    ↑                   ↑
# relation with other city_node  route_node(i-1)      route_node(i-1)     route_node(i-1)

class ExtendedNode:

    def __init__(self, extendend_node_id):
        """
        node to extended graph
        :param extendend_node_id: node id
        """
        self.id = extendend_node_id


class CityNode(ExtendedNode):

    def __init__(self, city_node_id, graph_node: Node):
        """
        extended node with graph node information
        :param city_node_id: node id
        :param graph_node: CBD, Subcenter or Periphery node
        """
        ExtendedNode.__init__(self, city_node_id)
        self.graph_node = graph_node


class StopNode(ExtendedNode):
    def __init__(self, stop_node_id, mode_obj: TransportMode, city_node: CityNode):
        """
        extended node with Transport Mode information
        :param stop_node_id: node id
        :param mode_obj: TransportMode object
        :param city_node: CityNode object associated
        """
        ExtendedNode.__init__(self, stop_node_id)
        self.mode = mode_obj
        self.city_node = city_node


class RouteNode(ExtendedNode):
    def __init__(self, route_node_id, route_obj: Route, direction: str, stop_node: StopNode, previous=None):
        """
        extended node with route information
        :param route_node_id: node id
        :param route_obj: Route object
        :param direction: "I" if RouteNode represents forward direction, "R" if RouteNode represents return direction
        :param stop_node: StopNode object associated
        :param previous: previous RouteNode in stop sequences of the route associated
        """
        ExtendedNode.__init__(self, route_node_id)
        self.route = route_obj
        self.direction = direction
        self.stop_node = stop_node
        self.prev_route_node = previous


class ExtendedEdge:
    def __init__(self, extended_edge_id, nodei, nodej, t, f, edge_type):
        """
        edge to extended graph
        :param extended_edge_id: edge id
        :param nodei: edge origin node
        :param nodej: edge detination node
        :param t: edge time in hours
        :param f: edge frequency in [veh/hr]
        :param edge_type: edge type (ExtendedEdgesType)
        """
        self.id = extended_edge_id
        self.nodei = nodei
        self.nodej = nodej
        self.t = t
        self.f = f
        self.type = edge_type


class ExtendedEdgesType(Enum):
    """
    extended edges type. ACCESS edge to edges between CityNode and StopNode and vice versa, Boarding edge to edges
    between StopNode and RouteNode, ALIGHTING edge to edges between RouteNode and StopNode, Route edge to edges
    between RouteNode and RouteNode
    """
    ACCESS = 1
    BOARDING = 2
    ALIGHTING = 3
    ROUTE = 4


defaultdict_float = defaultdict(float)
defaultdict2_route_node = defaultdict(lambda: defaultdict(List[RouteNode]))
defaultdict2_route_direction = defaultdict(lambda: defaultdict(List[Route, str]))


class ExtendedGraph:

    def __init__(self, graph_obj: Graph, routes: List[Route], sPTP: float, frequency_routes: defaultdict_float = None):
        """
        class to create extended graph that agglomerates the city graph information and transport routes in the network.
        It incorporates the penalty of the transfer of the passenger and the frequencies of the routes for
        the construction of edges
        :param graph_obj: Graph object
        :param routes: list of routes in network associated
        :param sPTP: penalty of the transfer of the passenger in EIV
        :param frequency_routes: defauldict(float) with key: route_id and value: frequency [veh/hr] of the route_id.
        Default value is a a dictionary with a frequency of 28 [veh/hr] for all routes
        """

        if frequency_routes is None:
            frequency_routes = defaultdict(float)

            for route in routes:
                frequency_routes[route.id] = route.mode.fini

        # list with all city_nodes
        city_nodes = self.build_city_nodes(graph_obj)

        # assistant dictionary to build stop_nodes and routes_nodes: dic[city_node][mode_obj] = [list of routes]
        tree_graph = self.build_tree_graph(routes, city_nodes)

        # list with all stop nodes, there are 1 stop_node for each mode transiting in a city_node
        stop_nodes = self.build_stop_nodes(tree_graph)

        # list with all route_nodes, there are 1 route_node connected to a stop_node for each route in a city_node
        route_nodes = self.build_route_nodes(routes, stop_nodes)

        # extended graph nodes like as a dictionary: dic[city_node][stop_node] = [list route_nodes]
        self.__extended_graph_nodes = self.build_extended_graph_nodes(route_nodes)

        # list with all access edges, edges between city_node<->stop_node
        access_edges = self.build_access_edges(self.__extended_graph_nodes)
        # list with all boarding edges, edges between stop_node->route_node
        boarding_edges = self.build_boarding_edges(self.__extended_graph_nodes, frequency_routes)
        # list with all alighting edges, edges between route_node->stop_node
        alighting_edges = self.build_alighting_edges(self.__extended_graph_nodes, sPTP)
        # list with all routes edges, edges between route_node(i-1)->route_node(i)
        routes_edges = self.build_route_edges(self.__extended_graph_nodes)

        self.__extended_graph_edges = []

        for edge in access_edges:
            self.__extended_graph_edges.append(edge)
        for edge in boarding_edges:
            self.__extended_graph_edges.append(edge)
        for edge in alighting_edges:
            self.__extended_graph_edges.append(edge)
        for edge in routes_edges:
            self.__extended_graph_edges.append(edge)

    def __str__(self):
        """
        to print extended graph
        :return: String with information of the extended graph
        """
        line = ""
        for city_node in self.__extended_graph_nodes:
            line += "City node\n-Graph node name: {}\n".format(city_node.graph_node.name)
            for stop_node in self.__extended_graph_nodes[city_node]:
                # information about access edge
                for edge in self.__extended_graph_edges:
                    if edge.nodei == city_node and edge.nodej == stop_node:
                        line += "\tAccess edge\n\t-Access time: {:.2f} [min]\n".format(edge.t)

                line += "\t\tStop node\n\t\t-Mode name: {}\n".format(stop_node.mode.name)

                for route_node in self.__extended_graph_nodes[city_node][stop_node]:
                    # information about boarding edge
                    for edge in self.__extended_graph_edges:
                        if edge.nodei == stop_node and edge.nodej == route_node:
                            line += "\t\t\tBoarding edge\n\t\t\t-Frequency: {:.2f} [veh/h]\n".format(edge.f)

                    # information about boarding edge
                    for edge in self.__extended_graph_edges:
                        if edge.nodei == route_node and edge.nodej == stop_node:
                            line += "\t\t\tAlighting edge\n\t\t\t-Penalty transfer: {:.2f} [min]\n".format(edge.t * 60)

                    # information about route node
                    if route_node.prev_route_node is None:
                        line += "\t\t\t\tRoute node\n\t\t\t\t-Route_id: {}\n\t\t\t\t-Direction: {}\n\t\t\t\t-Previous stop: {}\n\t\t\t\t-Time to previous stop: {:.2f} [hrs]\n".format(
                            route_node.route.id,
                            route_node.direction, "no data", 0)
                    else:
                        t = 0
                        for edge in self.__extended_graph_edges:
                            if edge.nodei == route_node.prev_route_node and edge.nodej == route_node:
                                t = edge.t
                                break
                        line += "\t\t\t\tRoute node\n\t\t\t\t-Route_id: {}\n\t\t\t\t-Direction: {}\n\t\t\t\t-Previous stop: {}\n\t\t\t\t-Time to previous stop: {} [hrs]\n".format(
                            route_node.route.id,
                            route_node.direction,
                            route_node.prev_route_node.stop_node.city_node.graph_node.name,
                            t)
        return line

    def get_extended_graph_nodes(self) -> defaultdict2_route_node:
        """
        to get extended nodes associated to the graph
        :return: dictionary: dic[CityNode][StopNode] = List[RouteNode]
        """
        return self.__extended_graph_nodes

    def get_extended_graph_edges(self) -> List[ExtendedEdge]:
        """
        to get extended edges associated to the graph
        :return: List[ExtendedEdges]
        """
        return self.__extended_graph_edges

    @staticmethod
    def build_city_nodes(graph_obj: Graph) -> List[CityNode]:
        """
        to build CityNodes in the extended graph
        :param graph_obj: Graph object
        :return: List[CityNode]
        """
        city_nodes = []
        for node in graph_obj.get_nodes():
            city_node = CityNode(len(city_nodes), node)
            city_nodes.append(city_node)
        return city_nodes

    @staticmethod
    def build_tree_graph(routes: List[Route], city_nodes: List[CityNode]) -> defaultdict2_route_direction:
        """
        to build auxiliary structure for construction of the extended graph
        :param routes: routes in the network associated
        :param city_nodes: List[CityNode]
        :return: dictionary: dic[CityNode][TransportMode] = List(Route, direction = "I" or "R")
        """

        tree_graph = defaultdict(lambda: defaultdict(list))
        for city_node in city_nodes:
            node_graph_id = city_node.graph_node.id

            for route in routes:
                mode_obj = route.mode
                stops_i = route.stops_sequence_i
                stops_r = route.stops_sequence_r

                for stop in stops_i:
                    if str(node_graph_id) == str(stop):
                        # to avoid add twice first and last node in circular routes
                        if route._type == RouteType.PREDEFINED or route._type == RouteType.CIRCULAR:
                            if (route, "I") not in tree_graph[city_node][mode_obj]:
                                tree_graph[city_node][mode_obj].append((route, "I"))
                        else:
                            tree_graph[city_node][mode_obj].append((route, "I"))

                for stop in stops_r:
                    if str(node_graph_id) == str(stop):
                        # to avoid add twice first and last node in circular routes
                        if route._type == RouteType.PREDEFINED or route._type == RouteType.CIRCULAR:
                            if (route, "R") not in tree_graph[city_node][mode_obj]:
                                tree_graph[city_node][mode_obj].append((route, "R"))
                        else:
                            tree_graph[city_node][mode_obj].append((route, "R"))
        return tree_graph

    @staticmethod
    def build_stop_nodes(tree_graph: defaultdict2_route_direction) -> List[StopNode]:
        """
        to build StopNode in extended graph
        :param tree_graph: dic[CityNode][TransportMode] = List(Route, direction = "I" or "R")
        :return: List[StopNode]
        """
        # list with tuples (mode_obj, city_node) to not duplicated more of a stop for each mode in a city_node
        mode_city = []
        stop_nodes = []
        for city_node in tree_graph:
            for mode_obj in tree_graph[city_node]:
                if (mode_obj, city_node) not in mode_city:
                    mode_city.append((mode_obj, city_node))
                    stop_node = StopNode(len(stop_nodes), mode_obj, city_node)
                    stop_nodes.append(stop_node)
        return stop_nodes

    @staticmethod
    def build_route_nodes(routes: List[Route], stop_nodes: List[StopNode]) -> List[RouteNode]:
        """
        to build route nodes in extended graph
        :param routes: list of routes in network associated
        :param stop_nodes: List[StopNode]
        :return: List[RouteNode]
        """
        # list with all route nodes
        route_nodes = []
        for route in routes:
            mode = route.mode
            stop_i = route.stops_sequence_i
            stop_r = route.stops_sequence_r
            # to circular exceptions
            _type = route._type

            # nodes for "I" direction
            nodes = []
            prev_route_node = None
            # look for a previous route node
            for stop in stop_i:
                stop_node = None
                for s in stop_nodes:
                    if s.mode == mode and str(s.city_node.graph_node.id) == str(stop):
                        stop_node = s
                        break
                route_node = RouteNode(len(route_nodes) + len(nodes), route, "I", stop_node, prev_route_node)
                prev_route_node = route_node
                nodes.append(route_node)
            # add previous node_route in circular routes
            if prev_route_node is not None and _type == RouteType.CIRCULAR:
                for i in range(len(nodes)):
                    if i == 0:
                        continue
                    n = nodes[i]
                    if i == 1:
                        n.prev_route_node = nodes[len(nodes) - 1]
                        route_nodes.append(n)
                        continue
                    n = nodes[i]
                    route_nodes.append(n)
            # add previous node_route
            if _type == RouteType.PREDEFINED or _type == RouteType.CUSTOM:
                for n in nodes:
                    route_nodes.append(n)
            # nodes for "R" direction
            nodes = []
            prev_route_node = None
            # look for a previous route node
            for stop in stop_r:
                stop_node = None
                for s in stop_nodes:
                    if s.mode == mode and str(s.city_node.graph_node.id) == str(stop):
                        stop_node = s
                        break
                route_node = RouteNode(len(route_nodes) + len(nodes), route, "R", stop_node, prev_route_node)
                prev_route_node = route_node
                nodes.append(route_node)
            # add previous node_route in circular routes
            if prev_route_node is not None and _type == RouteType.CIRCULAR:
                for i in range(len(nodes)):
                    if i == 0:
                        continue
                    n = nodes[i]
                    if i == 1:
                        n.prev_route_node = nodes[len(nodes) - 1]
                        route_nodes.append(n)
                        continue
                    n = nodes[i]
                    route_nodes.append(n)
            # add previous node_route
            if _type == RouteType.PREDEFINED or _type == RouteType.CUSTOM:
                for n in nodes:
                    route_nodes.append(n)

        return route_nodes

    @staticmethod
    def build_extended_graph_nodes(route_nodes: List[RouteNode]) -> defaultdict2_route_node:
        """
        to build extended graph node structure
        :param route_nodes: List[RouteNode]
        :return: dictionary: dic[CityNode][StopNode] = List[RouteNode]
        """

        extended_graph_nodes = defaultdict(lambda: defaultdict(list))
        for route_node in route_nodes:
            # print(route_node)
            stop_node = route_node.stop_node
            city_node = stop_node.city_node
            extended_graph_nodes[city_node][stop_node].append(route_node)

        return extended_graph_nodes

    @staticmethod
    def build_access_edges(extended_graph_nodes: defaultdict2_route_node) -> List[ExtendedEdge]:
        """
        to build access edges in extended graph
        :param extended_graph_nodes: dictionary: dic[CityNode][StopNode] = List[RouteNode]
        :return: List[ExtendedEdge]
        """
        access_edges = []
        for city_node in extended_graph_nodes:
            for stop_node in extended_graph_nodes[city_node]:
                edge1 = ExtendedEdge(len(access_edges), city_node, stop_node,
                                     stop_node.mode.tat / 60, float('inf'), ExtendedEdgesType.ACCESS)
                access_edges.append(edge1)
                edge2 = ExtendedEdge(len(access_edges), stop_node, city_node,
                                     stop_node.mode.tat / 60, float('inf'), ExtendedEdgesType.ACCESS)
                access_edges.append(edge2)
        return access_edges

    @staticmethod
    def build_boarding_edges(extended_graph_nodes: defaultdict2_route_node, frequency_routes: defaultdict_float) -> \
            List[ExtendedEdge]:
        """
        to build boarding edges in extended graph
        :param extended_graph_nodes: dic[CityNode][StopNode] = List[RouteNode]
        :param frequency_routes: defauldict(float) with key: route_id and value: frequency [veh/hr] of the route_id.
        :return: List[ExtendedEdge]
        """
        boarding_edges = []
        for city_node in extended_graph_nodes:
            for stop_node in extended_graph_nodes[city_node]:
                for route_node in extended_graph_nodes[city_node][stop_node]:
                    route = route_node.route
                    direction = route_node.direction

                    if direction == "I":
                        node_sequence = route.nodes_sequence_i
                    else:
                        node_sequence = route.nodes_sequence_r
                    # if not be last stop of the route add boarding edges
                    if route._type != RouteType.CIRCULAR and str(route_node.stop_node.city_node.graph_node.id) != str(
                            node_sequence[len(node_sequence) - 1]):
                        edge = ExtendedEdge(len(boarding_edges), stop_node, route_node,
                                            0, frequency_routes[route.id] / route.mode.d, ExtendedEdgesType.BOARDING)
                        boarding_edges.append(edge)
                        continue
                    # but if route type is circular add always boarding edges
                    if route._type == RouteType.CIRCULAR:
                        edge = ExtendedEdge(len(boarding_edges), stop_node, route_node,
                                            0, frequency_routes[route.id] / route.mode.d, ExtendedEdgesType.BOARDING)
                        boarding_edges.append(edge)
                        continue

        return boarding_edges

    @staticmethod
    def build_alighting_edges(extended_graph_nodes: defaultdict2_route_node, sPTP: float) -> List[ExtendedEdge]:
        """
        to build extended alighting edges
        :param extended_graph_nodes: dic[CityNode][StopNode] = List[RouteNode]
        :param sPTP: penalty of the transfer of the passenger in EIV
        :return: List[ExtendedEdge]
        """
        spt = sPTP / 60
        alighting_edges = []
        for city_node in extended_graph_nodes:
            for stop_node in extended_graph_nodes[city_node]:
                for route_node in extended_graph_nodes[city_node][stop_node]:
                    route = route_node.route
                    direction = route_node.direction
                    if direction == "I":
                        node_sequence = route.nodes_sequence_i
                    else:
                        node_sequence = route.nodes_sequence_r
                    # if not be first stop of the route add alighting edges
                    if route._type != RouteType.CIRCULAR and str(route_node.stop_node.city_node.graph_node.id) != str(
                            node_sequence[0]):
                        edge = ExtendedEdge(len(alighting_edges), route_node, stop_node,
                                            spt, float('inf'), ExtendedEdgesType.ALIGHTING)
                        alighting_edges.append(edge)
                        continue
                    # but if route type is circular add always alighting edges
                    if route._type == RouteType.CIRCULAR:
                        edge = ExtendedEdge(len(alighting_edges), route_node, stop_node,
                                            spt, float('inf'), ExtendedEdgesType.ALIGHTING)
                        alighting_edges.append(edge)
                        continue

        return alighting_edges

    @staticmethod
    def build_route_edges(extended_graph_nodes: defaultdict2_route_node) -> List[ExtendedEdge]:
        """
        to build extended routes edges
        :param extended_graph_nodes: dic[CityNode][StopNode] = List[RouteNode]
        :return: List[ExtendedEdge]
        """

        route_nodes = []
        nodes = []

        for city_node in extended_graph_nodes:
            nodes.append(city_node.graph_node)
            for stop_node in extended_graph_nodes[city_node]:
                for route_node in extended_graph_nodes[city_node][stop_node]:
                    route_nodes.append(route_node)

        route_edges = []
        for route_node in route_nodes:
            # only if route node has previous route node
            if route_node.prev_route_node is not None:

                # previous route node
                previous_route_node = route_node.prev_route_node

                # to identify route
                route = route_node.route
                direction = route_node.direction

                # velocity of the route
                v = route_node.stop_node.mode.v

                # id graph node to actual route node and previous route node
                id_city1 = route_node.stop_node.city_node.graph_node.id
                id_city2 = previous_route_node.stop_node.city_node.graph_node.id

                # to identify node sequence of the route
                if direction == "I":
                    node_sequence = route.nodes_sequence_i
                else:
                    node_sequence = route.nodes_sequence_r

                # to get distance between actual route node and previous route node
                distance = 0
                x_prev = previous_route_node.stop_node.city_node.graph_node.x
                y_prev = previous_route_node.stop_node.city_node.graph_node.y
                x_final = route_node.stop_node.city_node.graph_node.x
                y_final = route_node.stop_node.city_node.graph_node.y
                count = False
                for node in node_sequence:
                    if str(node) == str(id_city2):
                        count = True
                        continue
                    if str(node) == str(id_city1):
                        distance = distance + ((x_prev - x_final) ** 2 + (y_prev - y_final) ** 2) ** 0.5
                        break
                    if count:
                        for n in nodes:
                            if str(node) == str(n.id):
                                x = n.x
                                y = n.y
                                distance = distance + ((x_prev - x) ** 2 + (y_prev - y) ** 2) ** 0.5
                                x_prev = x
                                y_prev = y
                                break

                t = distance / v

                edge = ExtendedEdge(len(route_edges), previous_route_node, route_node,
                                    t, float('inf'), ExtendedEdgesType.ROUTE)
                route_edges.append(edge)
        return route_edges



from __future__ import annotations

import operator

from sidermit.exceptions import VaIsNotValidExceptions, \
    PvIsNotValidExceptions, PwIsNotValidExceptions, PaIsNotValidExceptions, \
    PtIsNotValidExceptions, SpvIsNotValidExceptions, SpwIsNotValidExceptions, \
    SpaIsNotValidExceptions, SptIsNotValidExceptions


def passenger_property(property_name, comp_function, exception_instance,
                       docstring=None):
    def getter(self):
        return self.__getattribute__(property_name)

    def setter(self, new_value):
        if new_value is None or comp_function(new_value, 0):
            raise exception_instance
        self.__setattr__(property_name, new_value)

    return property(getter, setter, doc=docstring)


class Passenger:
    va = passenger_property('_va', operator.le, VaIsNotValidExceptions(
        "You must give a value > 0 for va"), docstring="Walking speed [km/h]")
    pv = passenger_property('_pv', operator.lt, PvIsNotValidExceptions(
        "You must give a value >= 0 for pv"), docstring="Value of in-vehicle time savings [US$/h]")
    pw = passenger_property('_pw', operator.lt, PwIsNotValidExceptions(
        "You must give a value >= 0 for pw"), docstring="Value of waiting time savings [US$/h]")
    pa = passenger_property('_pa', operator.lt, PaIsNotValidExceptions(
        "You must give a value >= 0 for pa"), docstring="Value of access time savings [US$/h]")
    pt = passenger_property('_pt', operator.lt, PtIsNotValidExceptions(
        "You must give a value >= 0 for pt"), docstring="Pure transfer penalty [EIV]")
    spv = passenger_property('_spv', operator.lt, SpvIsNotValidExceptions(
        "You must give a value >= 0 for spv"), docstring="Subjetive value of in-vehicle time savings [US$/h]")
    spw = passenger_property('_spw', operator.lt, SpwIsNotValidExceptions(
        "You must give a value >= 0 for spw"), docstring="Subjetive value of waiting time savings [US$/h]")
    spa = passenger_property('_spa', operator.lt, SpaIsNotValidExceptions(
        "You must give a value >= 0 for spa"), docstring="Subjetive value of access time savings [US$/h]")
    spt = passenger_property('_spt', operator.lt, SptIsNotValidExceptions(
        "You must give a value >= 0 for spt"), docstring="Subjetive pure transfer penalty [EIV]")

    def __init__(self, va: float, pv: float, pw: float, pa: float, pt: float,
                 spv: float, spw: float, spa: float, spt: float):
        self.va = va
        self.pv = pv
        self.pw = pw
        self.pa = pa
        self.pt = pt
        self.spv = spv
        self.spw = spw
        self.spa = spa
        self.spt = spt

    @staticmethod
    def get_default_passenger() -> Passenger:
        """
        to get passenger with default values
        :return: Passenger
        """
        return Passenger(4.0, 2.74, 5.48, 8.22, 16, 2.74, 5.48, 8.22, 16)



from collections import defaultdict
from typing import List

from sidermit.city import Graph
from sidermit.publictransportsystem import Route, TransportMode

defaultdict_float = defaultdict(float)


class Constrains:

    @staticmethod
    def most_loaded_section_constrains(routes: List[Route], most_loaded_section: defaultdict_float) -> List[float]:
        """
        to get constrains to optimization problem with respect to most loaded section for each routes
        :param most_loaded_section: dict with most loaded section for each route_id
        :param routes: list of Route object
        :return: list with constrains
        """

        ineq_constrains = []

        for route in routes:
            kmax = route.mode.kmax
            max_loaded_section = most_loaded_section[route.id]

            ineq_constrains.append(max_loaded_section - kmax)

        return ineq_constrains

    @staticmethod
    def fmax_constrains(graph_obj: Graph, routes: List[Route], list_mode: List[TransportMode], f: defaultdict_float) -> \
            List[float]:
        """
        to get constrains about fmax in each edge in the network with respect to capacity in stop of the each mode
        :param graph_obj: Graph object
        :param routes: list of Route object
        :param list_mode: list TransportMode object
        :param f: dict with frequency for each route_id
        :return: list with constrains
        """

        ineq_constrain = []

        edges = graph_obj.get_edges()

        for edge in edges:
            nodei = edge.node1.id
            nodej = edge.node2.id
            for mode in list_mode:
                fmax = mode.fmax
                sum_f = 0
                for route in routes:
                    if route.mode == mode:
                        if f[route.id] != 0:
                            node_sequence_i = route.nodes_sequence_i
                            node_sequence_r = route.nodes_sequence_r

                            for i in range(len(node_sequence_i) - 1):
                                j = i + 1
                                if str(nodei) == str(node_sequence_i[i]) and str(nodej) == str(node_sequence_i[j]):
                                    sum_f += f[route.id] / mode.d

                            for i in range(len(node_sequence_r) - 1):
                                j = i + 1
                                if str(nodei) == str(node_sequence_r[i]) and str(nodej) == str(node_sequence_r[j]):
                                    sum_f += f[route.id] / mode.d
                ineq_constrain.append(sum_f - fmax)

        return ineq_constrain



from collections import defaultdict
from typing import List

import networkx as nx
from matplotlib import pyplot as plt

from sidermit.exceptions import *
from sidermit.optimization.preoptimization import ExtendedGraph, CityNode, StopNode, RouteNode, ExtendedEdgesType, \
    ExtendedEdge, ExtendedNode
from sidermit.publictransportsystem import Passenger, TransportModeManager

defaultdict2_float = defaultdict(lambda: defaultdict(float))
list_suc = defaultdict(List[ExtendedEdge])
list_lab = defaultdict(float)
list_f = defaultdict(float)
list_elemental_path = List[ExtendedNode]
defaultdict_elemental_path = defaultdict(List[list_elemental_path])

dic_hyperpaths = defaultdict(lambda: defaultdict(lambda: defaultdict(List[list_elemental_path])))
dic_labels = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_successors = defaultdict(lambda: defaultdict(lambda: defaultdict(List[ExtendedEdge])))
dic_frequency = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_Vij = defaultdict(lambda: defaultdict(float))


class Hyperpath:

    def __init__(self, extended_graph_obj: ExtendedGraph, passenger_obj: Passenger):
        """
        class to Hyperpath algorithm
        :param extended_graph_obj: ExtendedGraph object
        :param passenger_obj: Passenger object
        """
        self.extended_graph_obj = extended_graph_obj
        self.passenger_obj = passenger_obj

    def network_validator(self, OD_matrix: defaultdict2_float) -> bool:
        """
        to check if Transport network is well defined for all pairs OD with trips. This must has at least a route for
        each OD pair with trips. Also this must has until 2 TransportMode and at least one has parameter d=1.
        :param OD_matrix: OD matrix get from Demand object
        :return: True if all OD pairs with trips have at least one path between origin and destination. False if not.
        """
        nodes = self.extended_graph_obj.get_extended_graph_nodes()
        # to check a path between all OD pair with trips
        for origin_id in OD_matrix:
            for destination_id in OD_matrix[origin_id]:
                vij = OD_matrix[origin_id][destination_id]
                if vij != 0:
                    origin = None
                    destination = None
                    for city_node in nodes:
                        if str(origin_id) == str(city_node.graph_node.id):
                            origin = city_node
                        if str(destination_id) == str(city_node.graph_node.id):
                            destination = city_node

                    _, label, _ = self.build_hyperpath_graph(origin, destination)

                    # if there is a stop with a label != infinity, you can get from the origin to the destination
                    conection = False
                    for stop in nodes[origin]:
                        if label[stop] != float('inf'):
                            conection = True
                            break
                    if conection is False:
                        return False
        # to check network must has until 2 TransportMode
        list_mode = []
        for city_node in nodes:
            for stop in nodes[city_node]:
                if stop.mode not in list_mode:
                    list_mode.append(stop.mode)

        mode_manager = TransportModeManager(add_default_mode=False)

        for mode in list_mode:
            mode_manager.add_mode(mode)

        return mode_manager.is_valid_to_assignment_step()

    def build_hyperpath_graph(self, node_city_origin: CityNode, node_city_destination: CityNode) -> (
            list_suc, list_lab, list_f):
        """
        build the entire graph to connect the origin and destination with the hyperpath algorithm
        :param node_city_origin: origin CityNode
        :param node_city_destination: destination CityNode
        :return: successors , label, frequencies
        """

        nodes = self.extended_graph_obj.get_extended_graph_nodes()
        edges = self.extended_graph_obj.get_extended_graph_edges()

        # we initialize node labels at infinity except for the destination with label 0
        labels = defaultdict(float)
        labels_inf = defaultdict(float)
        # successors will be initialized empty
        successor = defaultdict(list)
        successor_inf = defaultdict(None)
        # frequency will be initialized to zero
        frequencies = defaultdict(float)
        # list of processed nodes (with calculated strategy)
        S = []

        # we initialize parameters
        # number of nodes in the extended graph
        n_nodes = 0
        for city_node in nodes:
            if city_node == node_city_destination:
                labels[city_node] = 0
                labels_inf[city_node] = 0
                frequencies[city_node] = 0
                n_nodes = n_nodes + 1
            else:
                labels[city_node] = float('inf')
                labels_inf[city_node] = float('inf')
                frequencies[city_node] = 0
                n_nodes = n_nodes + 1
            for stop_node in nodes[city_node]:
                labels[stop_node] = float('inf')
                labels_inf[stop_node] = float('inf')
                frequencies[stop_node] = 0
                n_nodes = n_nodes + 1
                for route_node in nodes[city_node][stop_node]:
                    labels[route_node] = float('inf')
                    labels_inf[route_node] = float('inf')
                    frequencies[route_node] = 0
                    n_nodes = n_nodes + 1

        # while there are nodes that have not been processed
        while len(S) != n_nodes:
            # we find node with minimum label and that does not belong to S
            min_label_node = None
            min_label = float('inf')
            for node in labels:
                if node not in S:
                    if min(labels[node], labels_inf[node]) < min_label:
                        min_label = min(labels[node], labels_inf[node])
                        min_label_node = node
            # node to be processed, initially equals destination
            j = min_label_node
            # update S
            S.append(j)
            # we must find all edges that end in j and whose beginning is not in S
            edge_j = []
            for edge in edges:
                # we will remove the edges of access to the CityNode of origin
                # because each StopNode in origin must have its own hyperpath
                if edge.nodei == node_city_origin:
                    continue
                if edge.nodej == node_city_origin:
                    continue
                if edge.nodei not in S and edge.nodej == j:
                    edge_j.append(edge)

            # for edges we update the label of the origin node as: labeli = labelj + time_arco ij
            for edge in edge_j:
                # not to consider transfer penalty at origin and include a penalty of access time
                edge_t = edge.t
                if edge.type == ExtendedEdgesType.ALIGHTING:
                    if edge.nodej.city_node == node_city_destination:
                        edge_t = 0
                if edge.type == ExtendedEdgesType.ACCESS:
                    edge_t = edge.t * self.passenger_obj.spa / self.passenger_obj.spv

                # equivalent to ~t_a
                t_i = edge_t + min(labels[j], labels_inf[j])
                i = edge.nodei

                # for all types of edges except boarding
                if edge.f == float('inf') and t_i < labels_inf[i]:
                    successor_inf[i] = edge
                    labels_inf[i] = t_i

                # for  boarding edges
                if edge.f < float('inf') and t_i < labels[i]:
                    theta = i.mode.theta

                    # initial case
                    if frequencies[i] == 0 and labels[i] == float('inf'):
                        # print(edge.type)
                        successor[i].append(edge)
                        labels[i] = (theta * self.passenger_obj.spw / self.passenger_obj.spv + edge.f * t_i) / edge.f
                        frequencies[i] = frequencies[i] + edge.f
                    # previously assigned label
                    else:
                        successor[i].append(edge)
                        labels[i] = (frequencies[i] * labels[i] + edge.f * t_i) / (frequencies[i] + edge.f)
                        frequencies[i] = frequencies[i] + edge.f

                # we verify that all the successors of i remain optimal
                for edge_b in successor[i]:
                    if edge_b == edge:
                        continue
                    # not to consider transfer penalty at origin and include a penalty of access time
                    edge_b_t = edge_b.t
                    if edge_b.type == ExtendedEdgesType.ALIGHTING:
                        if edge.nodej.city_node == node_city_destination:
                            edge_b_t = 0
                    if edge_b.type == ExtendedEdgesType.ACCESS:
                        edge_b_t = edge_b.t * self.passenger_obj.spa / self.passenger_obj.spv

                    # equivalent to ~t_b
                    t_ib = min(labels[edge_b.nodej], labels_inf[edge_b.nodej]) + edge_b_t

                    # remove sub optimal edge in the successors list
                    if t_ib >= labels[i]:
                        successor[i].remove(edge_b)
                        labels[i] = (frequencies[i] * labels[i] - edge_b.f * t_ib) / (frequencies[i] - edge_b.f)
                        frequencies[i] = frequencies[i] - edge_b.f

        # we reduce successor lists and labels to a single list
        successors = defaultdict(list)
        label = defaultdict(float)

        for city_node in nodes:
            if labels[city_node] < labels_inf[city_node]:
                label[city_node] = labels[city_node]
                for suc in successor[city_node]:
                    successors[city_node].append(suc)
            else:
                label[city_node] = labels_inf[city_node]

                if successor_inf.get(city_node):
                    successors[city_node].append(successor_inf[city_node])

            for stop_node in nodes[city_node]:
                if labels[stop_node] < labels_inf[stop_node]:
                    label[stop_node] = labels[stop_node]
                    if city_node == node_city_origin:
                        label[stop_node] = labels[
                                               stop_node] + stop_node.mode.tat / 60 * self.passenger_obj.spa / self.passenger_obj.spv
                    for suc in successor[stop_node]:
                        successors[stop_node].append(suc)
                else:
                    label[stop_node] = labels_inf[stop_node]
                    if city_node == node_city_origin:
                        label[stop_node] = labels_inf[
                                               stop_node] + stop_node.mode.tat / 60 * self.passenger_obj.spa / self.passenger_obj.spv

                    if successor_inf.get(stop_node):
                        successors[stop_node].append(successor_inf[stop_node])
                for route_node in nodes[city_node][stop_node]:
                    if labels[route_node] < labels_inf[route_node]:
                        label[route_node] = labels[route_node]
                        for suc in successor[route_node]:
                            successors[route_node].append(suc)
                    else:
                        label[route_node] = labels_inf[route_node]
                        if successor_inf.get(route_node):
                            successors[route_node].append(successor_inf[route_node])

        return successors, label, frequencies

    @staticmethod
    def string_hyperpath_graph(successors: list_suc, label: list_lab, frequencies: list_f) -> str:
        """
        String with the representation of the hyperpath graph
        :param successors: dic[ExtendedNode] = List[ExtendedNode]. Dictionary that gives the relation of successor
        nodes in the hyperpath
        :param label: dic[ExtendedNode] = label. Dictionary that gives the label for each ExtendedNode in hours
        with weight equivalent to the travel time
        :param frequencies: dic[ExtendedNode] = frequency. Dictionary that gives the cumulative frequency of all
        successors for each ExtendedNode
        :return: String with the representation of the hyperpath graph
        """
        line = "HyperPath Graph\n"
        for node in label:

            line_frequency = frequencies[node]

            line_successor = ""

            for suc in successors[node]:
                nodei = suc.nodei
                nodej = suc.nodej
                linei = ""
                linej = ""
                if isinstance(nodei, CityNode):
                    linei += "Citynode: {}".format(nodei.graph_node.name)
                if isinstance(nodei, StopNode):
                    linei += "Stopnode {}: {}".format(nodei.mode.name, nodei.city_node.graph_node.name)
                if isinstance(nodei, RouteNode):
                    linei += "Routenode {} {}: {}".format(nodei.route.id, nodei.direction,
                                                          nodei.stop_node.city_node.graph_node.name)
                if isinstance(nodej, CityNode):
                    linej += "Citynode: {}".format(nodej.graph_node.name)
                if isinstance(nodej, StopNode):
                    linej += "Stopnode {}: {}".format(nodej.mode.name, nodej.city_node.graph_node.name)
                if isinstance(nodej, RouteNode):
                    linej += "Routenode {} {}: {}".format(nodej.route.id, nodej.direction,
                                                          nodej.stop_node.city_node.graph_node.name)

                line_successor += "[{}: {} - {}] ".format(suc.type, linei, linej)

            if isinstance(node, CityNode):
                line += "City_node\n\t-Graph_node_name: {}\n\t-label: {:.2f}\n\t-Successor: {}\n\t-Frequencies: {:.2f}\n".format(
                    node.graph_node.name, label[node], line_successor, line_frequency)
            if isinstance(node, StopNode):
                line += "Stop_node\n\t-Mode_name: {}\n\t-Graph_node_name: {}\n\t-label: {:.2f}\n\t-Successor: {}\n\t-Frequencies: {:.2f}\n".format(
                    node.mode.name, node.city_node.graph_node.name, label[node], line_successor, line_frequency)
            if isinstance(node, RouteNode):
                line += "Route_node\n\t-route_id: {}\n\t-direction: {}\n\t-Graph_node_name: {}\n\t-label: {:.2f}\n\t-Successor: {}\n\t-Frequencies: {:.2f}\n".format(
                    node.route.id, node.direction, node.stop_node.city_node.graph_node.name,
                    label[node], line_successor, line_frequency)
        return line

    def get_hyperpath_OD(self, origin: CityNode, destination: CityNode) -> (
            defaultdict_elemental_path, list_lab, list_suc):
        """
        to get all elemental path for each StopNode in Origin
        :param origin: CityNode origin
        :param destination: CityNode destination
        :return: (Dic[TransportMode] = List[List[ExtendedNodes]], dic[ExtendedNode] = Label, dic[ExtendedNode] =
        List[ExtendedEdge]). Each List[ExtendedNodes] represent a elemental path to connect origin and destination.
        List[ExtendedEdge] represent all successors edge for each ExtendedNode.
        """
        # we run hyperpath algorithm
        successors, label, frequencies = self.build_hyperpath_graph(origin, destination)

        nodes = self.extended_graph_obj.get_extended_graph_nodes()

        # dictionary with key: TransportMode and value all elemental path associated
        hyperpaths_od = defaultdict(list)

        # for each StopNode in Origin
        for stop in nodes[origin]:
            # hyperpath in the StopNode
            hyperpath_stop = [[origin, stop]]
            # we initialize hyperpath

            while True:
                # stop condition that each elemental path has reached the destination
                end = True
                for path in hyperpath_stop:
                    # there is a elemental path that has not reached the destination
                    if path[len(path) - 1] != destination:
                        end = False
                        break
                if end:
                    break

                new_hyperpath_stop = []
                # we add successors of those paths that have not reached the destination
                for path in hyperpath_stop:
                    # elemental path that has not reached the destination
                    if path[len(path) - 1] != destination:
                        # we add new elemental path as successors have the last node of the path analyzed
                        for suc in successors[path[len(path) - 1]]:
                            new_path = []
                            new_path.extend(path)
                            new_path.append(suc.nodej)
                            new_hyperpath_stop.append(new_path)
                    # path that arrived at destination
                    else:
                        new_hyperpath_stop.append(path)
                hyperpath_stop = new_hyperpath_stop

            for elemental_path in hyperpath_stop:
                hyperpaths_od[stop].append(elemental_path)

        return hyperpaths_od, label, successors, frequencies

    @staticmethod
    def string_hyperpaths_OD(hyperpaths_od: defaultdict_elemental_path, label: list_lab) -> str:
        """
        String with the representation of the hyperpath for a OD pair
        :param hyperpaths_od: Dic[TransportMode] = List[List[ExtendedNodes]]. Each List[ExtendedNodes] represent a elemental
        path to connect a origin and destination.
        :param label: dic[ExtendedNode] = label. Dictionary that gives the label for each ExtendedNode in hours
        with weight equivalent to the travel time
        :return: String with the representation of the hyperpath for a OD pair
        """
        line = ""
        for stop in hyperpaths_od:
            line += "\n{} stop\n".format(stop.mode.name)
            for path in hyperpaths_od[stop]:
                line += "\n\tNew Path:\n\t\t"
                for node in path:
                    if isinstance(node, CityNode):
                        line += "[City_node {}: {:.4f}]\n\t\t".format(node.graph_node.name, label[node])
                    if isinstance(node, StopNode):
                        line += "[Stop_node {} - {}: {:.4f}]\n\t\t".format(node.mode.name,
                                                                           node.city_node.graph_node.name, label[node])
                    if isinstance(node, RouteNode):
                        line += "[Route_node {} {} - {}: {:.4f}]\n\t\t".format(node.route.id, node.direction,
                                                                               node.stop_node.city_node.graph_node.name,
                                                                               label[node])
        return line

    def get_all_hyperpaths(self, OD_matrix: defaultdict2_float) -> (
            dic_hyperpaths, dic_labels, dic_successors, dic_frequency, dic_Vij):
        """
        get information about all hyperpath and label for all OD pair with trips in OD matrix
        :param OD_matrix:  OD matrix get from Demand object
        :return: (Dic[origin: CityNode][destination: CityNode][StopNode] = List[List[ExtendedNodes]],
        dic[origin: CityNode][destination: CityNode][ExtendedNode] = Label, dic[origin: CityNode][destination: CityNode]
        [ExtendedNode] = List[ExtendedEdge], dic[origin: CityNode][destination: CityNode][ExtendedNode] = float [veh/hr]
        , dic[origin][destination] = vij). Each List[ExtendedNodes] represent a elemental path to connect a origin
        and destination. List[ExtendedEdge] represent all successors edge for each ExtendedNode in a OD pair.
        """
        hyperpaths = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        labels = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
        successors = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        frequency = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
        Vij = defaultdict(lambda: defaultdict(list))

        nodes = self.extended_graph_obj.get_extended_graph_nodes()

        if self.network_validator(OD_matrix):
            for origin_id in OD_matrix:
                for destination_id in OD_matrix[origin_id]:
                    vij = OD_matrix[origin_id][destination_id]
                    if vij != 0:
                        origin = None
                        destination = None
                        for city_node in nodes:
                            if str(origin_id) == str(city_node.graph_node.id):
                                origin = city_node
                            if str(destination_id) == str(city_node.graph_node.id):
                                destination = city_node

                        hyperpaths_od, label, successor, frequencies = self.get_hyperpath_OD(origin, destination)

                        for city_node in nodes:
                            labels[origin][destination][city_node] = label[city_node]
                            for stop_node in nodes[city_node]:
                                labels[origin][destination][stop_node] = label[stop_node]
                                for route_node in nodes[city_node][stop_node]:
                                    labels[origin][destination][route_node] = label[route_node]

                        for city_node in nodes:
                            for suc in successor[city_node]:
                                successors[origin][destination][city_node].append(suc)
                            for stop_node in nodes[city_node]:
                                for suc in successor[stop_node]:
                                    successors[origin][destination][stop_node].append(suc)
                                for route_node in nodes[city_node][stop_node]:
                                    for suc in successor[route_node]:
                                        successors[origin][destination][route_node].append(suc)

                        for city_node in nodes:
                            frequency[origin][destination][city_node] = frequencies[city_node]
                            for stop_node in nodes[city_node]:
                                frequency[origin][destination][stop_node] = frequencies[stop_node]
                                for route_node in nodes[city_node][stop_node]:
                                    frequency[origin][destination][route_node] = frequencies[route_node]

                        for stop in hyperpaths_od:
                            for elemental_path in hyperpaths_od[stop]:
                                hyperpaths[origin][destination][stop].append(elemental_path)

                        Vij[origin][destination] = vij

        else:
            raise TransportNetworkIsNotValidException(
                "each OD pair with trips must have at least one path between origin and destination and no more of 2 TransportMode defined in the network")

        return hyperpaths, labels, successors, frequency, Vij

    @staticmethod
    def string_all_hyperpaths(hyperpaths: dic_hyperpaths, labels: dic_labels, successors: dic_successors,
                              vij: dic_Vij) -> str:
        """
        to get a string with a summary of the all hyperpaths for all OD pair with trips.
        :param hyperpaths: Dic[origin: CityNode][destination: CityNode][StopNode] = List[List[ExtendedNodes]]
        :param labels: dic[origin: CityNode][destination: CityNode][ExtendedNode] = Label
        :param successors: dic[origin: CityNode][destination: CityNode][ExtendedNode] = List[ExtendedEdge]
        :param vij: dic[origin: CityNode][destination: CityNode] = vij
        :return: string with a summary of the all hyperpaths for all OD pair with trips.
        """

        line = "\n"

        for origin in hyperpaths:
            for destination in hyperpaths[origin]:
                for stop in hyperpaths[origin][destination]:
                    line += "origin: {}, destination: {}, vij: {:.2f}\n\t mode: {}, label: {:.2f} [EIV], n° elemental paths: {}, n° elemental paths (successors): {}\n".format(
                        origin.graph_node.name,
                        destination.graph_node.name,
                        vij[origin][destination],
                        stop.mode.name, labels[origin][destination][stop],
                        len(hyperpaths[origin][
                                destination][stop]), len(successors[origin][destination][stop]))

        return line

    @staticmethod
    def plot(hyperpaths_od: defaultdict_elemental_path):
        """
        plot alls hyperpaths for a OD pair
        :param hyperpaths_od: Dic[TransportMode] = List[List[ExtendedNodes]]. Each List[ExtendedNodes] represent a
        elemental path to connect origin and destination.
        :return:
        """
        city_nodes = []
        stop_nodes = []
        route_nodes = []
        edges_graph = []
        for stop in hyperpaths_od:
            for path in hyperpaths_od[stop]:
                prev_node = None
                for node in path:
                    if isinstance(node, CityNode):
                        city_nodes.append(node.graph_node.name)
                        if prev_node is None:
                            prev_node = node.graph_node.name
                        else:
                            edges_graph.append((prev_node, node.graph_node.name))
                            prev_node = node.graph_node.name
                    if isinstance(node, StopNode):
                        stop_nodes.append("{} - {}".format(node.mode.name, node.city_node.graph_node.name))
                        if prev_node is None:
                            prev_node = "{} - {}".format(node.mode.name, node.city_node.graph_node.name)
                        else:
                            edges_graph.append(
                                (prev_node, "{} - {}".format(node.mode.name, node.city_node.graph_node.name)))
                            prev_node = "{} - {}".format(node.mode.name, node.city_node.graph_node.name)

                    if isinstance(node, RouteNode):
                        route_nodes.append("{} {} - {}".format(node.route.id, node.direction,
                                                               node.stop_node.city_node.graph_node.name))
                        if prev_node is None:
                            prev_node = "{} {} - {}".format(node.route.id, node.direction,
                                                            node.stop_node.city_node.graph_node.name)
                        else:
                            edges_graph.append((prev_node, "{} {} - {}".format(node.route.id, node.direction,
                                                                               node.stop_node.city_node.graph_node.name)))
                            prev_node = "{} {} - {}".format(node.route.id, node.direction,
                                                            node.stop_node.city_node.graph_node.name)

        G = nx.DiGraph()
        G.add_nodes_from(city_nodes)
        G.add_nodes_from(stop_nodes)
        G.add_nodes_from(route_nodes)
        G.add_edges_from(edges_graph)
        pos = nx.spring_layout(G, scale=10)  # positions for all nodes

        # nx.draw(G)
        # # separate calls to draw labels, nodes and edges
        # # plot p, Sc and CBD
        nx.draw_networkx_nodes(G, pos, cmap=plt.get_cmap('Set2'), nodelist=city_nodes, node_color='yellow',
                               node_size=200)
        nx.draw_networkx_nodes(G, pos, cmap=plt.get_cmap('Set2'), nodelist=stop_nodes, node_color='red',
                               node_size=200)
        nx.draw_networkx_nodes(G, pos, cmap=plt.get_cmap('Set2'), nodelist=route_nodes, node_color='green',
                               node_size=200)
        # # plot labels
        nx.draw_networkx_labels(G, pos, font_size=6)
        # # plot edges city
        nx.draw_networkx_edges(G, pos, edgelist=edges_graph, edge_color='orange', arrows=True)

        plt.title("City graph")
        plt.xlabel("X")
        plt.ylabel("Y")
        plt.gca().set_aspect('equal')
        plt.show()
        # plt.savefig(file_path)



import unittest
from collections import defaultdict

from sidermit.city import graph
from sidermit.optimization.preoptimization import ExtendedGraph, ExtendedEdgesType
from sidermit.publictransportsystem import *


class test_extended_graph(unittest.TestCase):

    def test_build_tree_graph(self):
        """
        test methods get_tree_graph
        :return:
        """
        graph_obj = graph.Graph.build_from_parameters(n=5, l=1000, g=0.5, p=2)
        network = TransportNetwork(graph_obj)

        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        metro_obj = mode_manager.get_mode("metro")
        # train_obj = TransportMode("train", 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

        feeder_routes = network.get_feeder_routes(bus_obj)
        radial_routes = network.get_radial_routes(metro_obj)
        circular_routes = network.get_circular_routes(bus_obj)

        for route in feeder_routes:
            network.add_route(route)

        for route in radial_routes:
            network.add_route(route)

        for route in circular_routes:
            network.add_route(route)

        city_nodes = ExtendedGraph.build_city_nodes(graph_obj)
        tree_graph = ExtendedGraph.build_tree_graph(network.get_routes(), city_nodes)

        one_stop = []
        two_stop = []
        three_stop = []

        four_route_city = 0
        six_route_city = 0
        ten_route_city = 0

        one_route_stop = 0
        two_route_stop = 0
        ten_route_stop = 0

        for city_node in tree_graph:
            if len(tree_graph[city_node]) == 1:
                one_stop.append(city_node)
            if len(tree_graph[city_node]) == 2:
                two_stop.append(city_node)
            if len(tree_graph[city_node]) == 3:
                three_stop.append(city_node)

            n_routes_city = 0
            for stop in tree_graph[city_node]:
                n_routes_stop = 0
                for _ in tree_graph[city_node][stop]:
                    n_routes_stop = n_routes_stop + 1
                    n_routes_city = n_routes_city + 1

                if n_routes_stop == 1:
                    one_route_stop = one_route_stop + 1
                if n_routes_stop == 2:
                    two_route_stop = two_route_stop + 1
                if n_routes_stop == 10:
                    ten_route_stop = ten_route_stop + 1

            if n_routes_city == 4:
                four_route_city = four_route_city + 1
            if n_routes_city == 6:
                six_route_city = six_route_city + 1
            if n_routes_city == 10:
                ten_route_city = ten_route_city + 1

        self.assertEqual(len(one_stop), 1)
        self.assertEqual(len(two_stop), 10)
        self.assertEqual(len(three_stop), 0)

        self.assertEqual(four_route_city, 5)
        self.assertEqual(six_route_city, 5)
        self.assertEqual(ten_route_city, 1)

        self.assertEqual(one_route_stop, 0)
        self.assertEqual(two_route_stop, 15)
        self.assertEqual(ten_route_stop, 1)

    def test_build_nodes(self):

        graph_obj = graph.Graph.build_from_parameters(n=5, l=1000, g=0.5, p=2)
        network = TransportNetwork(graph_obj)

        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        metro_obj = mode_manager.get_mode("metro")
        # train_obj = TransportMode("train", 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

        feeder_routes = network.get_feeder_routes(bus_obj)
        radial_routes = network.get_radial_routes(metro_obj)
        circular_routes = network.get_circular_routes(bus_obj)

        for route in feeder_routes:
            network.add_route(route)

        for route in radial_routes:
            network.add_route(route)

        for route in circular_routes:
            network.add_route(route)

        city_nodes = ExtendedGraph.build_city_nodes(graph_obj)
        tree_graph = ExtendedGraph.build_tree_graph(network.get_routes(), city_nodes)
        stop_nodes = ExtendedGraph.build_stop_nodes(tree_graph)
        route_nodes = ExtendedGraph.build_route_nodes(network.get_routes(), stop_nodes)

        self.assertEqual(len(city_nodes), 11)
        self.assertEqual(len(stop_nodes), 21)
        self.assertEqual(len(route_nodes), 60)

    def test_extended_graph_nodes(self):

        graph_obj = graph.Graph.build_from_parameters(n=5, l=1000, g=0.5, p=2)
        network = TransportNetwork(graph_obj)

        passenger_obj = Passenger(2, 2, 2, 2, 2, 2, 2, 2, 2)

        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        metro_obj = mode_manager.get_mode("metro")
        # train_obj = TransportMode("train", 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

        feeder_routes = network.get_feeder_routes(bus_obj)
        radial_routes = network.get_radial_routes(metro_obj)
        circular_routes = network.get_circular_routes(bus_obj)

        for route in feeder_routes:
            network.add_route(route)

        for route in radial_routes:
            network.add_route(route)

        for route in circular_routes:
            network.add_route(route)

        extended_graph = ExtendedGraph(graph_obj, network.get_routes(), passenger_obj.spt)
        extended_graph_nodes = extended_graph.get_extended_graph_nodes()

        n_city = 0
        n_stop = 0
        n_route = 0
        for city_node in extended_graph_nodes:
            n_city = n_city + 1
            for stop_node in extended_graph_nodes[city_node]:
                n_stop = n_stop + 1
                for _ in extended_graph_nodes[city_node][stop_node]:
                    n_route = n_route + 1

        self.assertEqual(n_city, 11)
        self.assertEqual(n_stop, 21)
        self.assertEqual(n_route, 60)

    def test_build_edges(self):

        graph_obj = graph.Graph.build_from_parameters(n=5, l=1000, g=0.5, p=2)
        network = TransportNetwork(graph_obj)

        passenger_obj = Passenger(4, 2, 2, 2, 2, 2, 2, 2, 2)

        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        metro_obj = mode_manager.get_mode("metro")
        # train_obj = TransportMode("train", 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

        feeder_routes = network.get_feeder_routes(bus_obj)
        radial_routes = network.get_radial_routes(metro_obj)
        circular_routes = network.get_circular_routes(bus_obj)

        for route in feeder_routes:
            network.add_route(route)

        for route in radial_routes:
            network.add_route(route)

        for route in circular_routes:
            network.add_route(route)

        city_nodes = ExtendedGraph.build_city_nodes(graph_obj)
        tree_graph = ExtendedGraph.build_tree_graph(network.get_routes(), city_nodes)
        stop_nodes = ExtendedGraph.build_stop_nodes(tree_graph)
        route_nodes = ExtendedGraph.build_route_nodes(network.get_routes(), stop_nodes)

        extended_graph_nodes = ExtendedGraph.build_extended_graph_nodes(route_nodes)

        initial_frequency = defaultdict(float)

        for route in network.get_routes():
            initial_frequency[route.id] = 28

        access_edges = ExtendedGraph.build_access_edges(extended_graph_nodes)
        boarding_edges = ExtendedGraph.build_boarding_edges(extended_graph_nodes, initial_frequency)
        alighting_edges = ExtendedGraph.build_alighting_edges(extended_graph_nodes, passenger_obj.spt)
        route_edges = ExtendedGraph.build_route_edges(extended_graph_nodes)

        self.assertEqual(len(access_edges), 42)
        self.assertEqual(len(boarding_edges), 40)
        self.assertEqual(len(alighting_edges), 40)
        self.assertEqual(len(route_edges), 40)

    def test_get_extended_edges(self):

        graph_obj = graph.Graph.build_from_parameters(n=5, l=1000, g=0.5, p=2)
        network = TransportNetwork(graph_obj)

        passenger_obj = Passenger(4, 2, 2, 2, 2, 2, 2, 2, 2)

        mode_manager = TransportModeManager()
        bus_obj = mode_manager.get_mode("bus")
        metro_obj = mode_manager.get_mode("metro")
        # train_obj = TransportMode("train", 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1)

        feeder_routes = network.get_feeder_routes(bus_obj)
        radial_routes = network.get_radial_routes(metro_obj, express=True)
        circular_routes = network.get_circular_routes(bus_obj)

        for route in feeder_routes:
            network.add_route(route)

        for route in radial_routes:
            network.add_route(route)

        for route in circular_routes:
            network.add_route(route)

        extended_graph = ExtendedGraph(graph_obj, network.get_routes(), passenger_obj.spt)
        extended_graph_edges = extended_graph.get_extended_graph_edges()

        n_access = 0
        n_boarding = 0
        n_alighting = 0
        n_route = 0

        for edge in extended_graph_edges:
            if edge.type == ExtendedEdgesType.ACCESS:
                n_access = n_access + 1
            if edge.type == ExtendedEdgesType.BOARDING:
                n_boarding = n_boarding + 1
            if edge.type == ExtendedEdgesType.ALIGHTING:
                n_alighting = n_alighting + 1
            if edge.type == ExtendedEdgesType.ROUTE:
                n_route = n_route + 1

        self.assertEqual(n_access, 32)
        self.assertEqual(n_boarding, 30)
        self.assertEqual(n_alighting, 30)
        self.assertEqual(n_route, 30)



from collections import defaultdict
from typing import List

from sidermit.optimization.preoptimization import RouteNode, StopNode, ExtendedGraph, CityNode, ExtendedEdge, \
    ExtendedNode
from sidermit.publictransportsystem import Passenger

defaultdict_float = defaultdict(float)
defaultdict2_float = defaultdict(lambda: defaultdict(float))
defaultdict3_float = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
list_elemental_path = List[ExtendedNode]
defaultdict_elemental_path = defaultdict(List[list_elemental_path])

dic_hyperpaths = defaultdict(lambda: defaultdict(lambda: defaultdict(List[list_elemental_path])))
dic_labels = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_successors = defaultdict(lambda: defaultdict(lambda: defaultdict(List[ExtendedEdge])))
dic_frequency = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_Vij = defaultdict(lambda: defaultdict(float))
dic_assigment = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_f = defaultdict(float)

dic_boarding = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))
dic_alighting = defaultdict(lambda: defaultdict(lambda: defaultdict(float)))

dic_loaded_section = defaultdict(float)


class UsersCost:

    @staticmethod
    def resources_consumer(hyperpaths: dic_hyperpaths, Vij: dic_Vij, assignment: dic_assigment,
                           successors: dic_successors, extended_graph: ExtendedGraph, vp: float, f: defaultdict_float,
                           z: defaultdict2_float, v: defaultdict3_float) -> (float, float, float, int):
        """
        to get resources consumer for all passenger in transport network access time, waiting time,
        time on board of vehicle, numbers of transfer
        :param hyperpaths: Dic[origin: CityNode][destination: CityNode][StopNode] = List[List[ExtendedNodes]]
        :param Vij: dic[origin: CityNode][destination: CityNode] = vij
        :param assignment: dic[origin: CityNode][destination: CityNode][Stop: StopNode] = %V_OD
        :param successors: dic[origin: CityNode][destination: CityNode][ExtendedNode] = List[ExtendedEdge]
        :param extended_graph: ExtendedGraph object
        :param vp: passenger velocity [km/hr]
        :param f: dict with frequency [veh/hr] for each route_id
        :param z: boarding, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param v: alighting, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :return: (access time, waiting time, time on boad, numbers of transfers): (float, float, float, int)
        """
        ta = 0
        tv = 0
        te = 0
        t = 0

        edges = extended_graph.get_extended_graph_edges()

        for origin in hyperpaths:
            for destination in hyperpaths[origin]:
                # viajes del par OD
                vod = Vij[origin][destination]
                for stop in hyperpaths[origin][destination]:
                    # viajes de todas las rutas elementales que salen de esta parada
                    vod_s = vod * assignment[origin][destination][stop] / 100

                    if vod_s == 0:
                        continue

                    paths = []

                    for suc in successors[origin][destination][stop]:
                        nodej = suc.nodej
                        paths.append((stop, nodej, vod_s))

                    # reportar ta inicial (lateral y tecnologico)
                    ta += vod_s * (stop.mode.tat / 60 + assignment[origin][destination][
                        stop] / 100 * origin.graph_node.width / (4 * vp * stop.mode.d))

                    while len(paths) != 0:
                        nodei, nodej, pax = paths.pop(0)

                        # evita continuar si llegaste a destino
                        if isinstance(nodei, StopNode) and nodei.city_node == destination:
                            continue

                        dis_pax = pax

                        # reportar te
                        if isinstance(nodei, StopNode):
                            if isinstance(nodej, RouteNode):

                                f_acum = 0

                                for suc in successors[origin][destination][nodei]:
                                    f_acum += f[suc.nodej.route.id]

                                dis_pax = dis_pax * (f[nodej.route.id] / f_acum)

                                te += dis_pax * nodei.mode.theta / (f_acum / nodej.route.mode.d)

                        # reportar tv
                        if isinstance(nodei, RouteNode):
                            if isinstance(nodej, RouteNode):
                                for edge in edges:
                                    if edge.nodei == nodei and edge.nodej == nodej:
                                        tv += dis_pax * edge.t
                                        break

                        # reportar transbordos y tv de bajada
                        if isinstance(nodei, RouteNode):
                            if isinstance(nodej, StopNode):
                                # transbordos
                                if nodej.city_node != destination:
                                    t += dis_pax

                                # para tiempo de viaje adicional por esperar la bajada del vehiculo
                                nodei_id = nodei.prev_route_node.stop_node.city_node.graph_node.id
                                nodej_id = nodei.stop_node.city_node.graph_node.id

                                stop_sequence_i = nodei.route.stops_sequence_i

                                index = 0
                                index_i = 0
                                index_j = 0

                                for node_id in stop_sequence_i:
                                    if str(node_id) == str(nodei_id):
                                        index_i = index
                                    if str(node_id) == str(nodej_id):
                                        index_j = index
                                    index = index + 1

                                tb = nodei.route.mode.t / 3600
                                # verificamos si es del sentido de ida
                                if index_i < index_j:
                                    pax_b = v[nodei.route.id]["I"][nodej]
                                    tv += (pax_b * 0.5 * tb) * dis_pax

                                # sentido de vuelta
                                else:
                                    pax_b = v[nodei.route.id]["R"][nodej]
                                    tv += (pax_b * 0.5 * tb) * dis_pax

                        # reportar ta
                        if isinstance(nodei, StopNode):
                            if isinstance(nodej, CityNode):
                                ta += nodei.mode.tat / 60
                        if isinstance(nodej, StopNode):
                            if isinstance(nodei, CityNode):
                                ta += nodej.mode.tat / 60

                        # agregamos elementos faltantes a las rutas
                        # si nodo i, nodo j y nodo k (new suc ) son de rutas entonces
                        for suc in successors[origin][destination][nodej]:
                            # agrega elementos para continuar rutas elementales
                            paths.append((nodej, suc.nodej, dis_pax))

                            # sumaremos tv producido por la espera de los que se bajan en paraderos cuando el ind sigue
                            # en ruta
                            # cbd                                 #SC                             #p
                            if isinstance(nodei, RouteNode) and isinstance(nodej, RouteNode) and isinstance(suc.nodej,
                                                                                                            RouteNode):
                                bya = nodei.route.mode.bya
                                tb = nodei.route.mode.t / 3600
                                # determinamos dirección
                                nodei_id = nodei.stop_node.city_node.graph_node.id
                                nodej_id = nodej.stop_node.city_node.graph_node.id

                                stop_sequence_i = nodei.route.stops_sequence_i

                                index = 0
                                index_i = 0
                                index_j = 0

                                for node_id in stop_sequence_i:
                                    if str(node_id) == str(nodei_id):
                                        index_i = index
                                    if str(node_id) == str(nodej_id):
                                        index_j = index
                                    index = index + 1

                                # verificamos si es del sentido de ida
                                if index_i < index_j:
                                    # simultaneo
                                    if bya == 1:
                                        pasajeros = max(z[nodei.route.id]["I"][nodej.stop_node],
                                                        v[nodei.route.id]["I"][nodej.stop_node])
                                        tv += pasajeros * tb * dis_pax
                                    # secuencial
                                    if bya == 0:
                                        pasajeros = z[nodei.route.id]["I"][nodej.stop_node] + v[nodei.route.id]["I"][
                                            nodej.stop_node]
                                        tv += pasajeros * tb * dis_pax
                                # sentido de vuelta
                                else:
                                    # simultaneo
                                    if bya == 1:
                                        pasajeros = max(z[nodei.route.id]["R"][nodej.stop_node],
                                                        v[nodei.route.id]["R"][nodej.stop_node])
                                        tv += pasajeros * tb * dis_pax
                                    # secuencial
                                    if bya == 0:
                                        pasajeros = z[nodei.route.id]["R"][nodej.stop_node] + v[nodei.route.id]["R"][
                                            nodej.stop_node]
                                        tv += pasajeros * tb * dis_pax

        return ta, te, tv, t

    def get_users_cost(self, hyperpaths: dic_hyperpaths, Vij: dic_Vij, assignment: dic_assigment,
                       successors: dic_successors, extended_graph: ExtendedGraph, f: defaultdict_float,
                       passenger_obj: Passenger, z: defaultdict3_float, v: defaultdict3_float) -> float:
        """
        to get users cost
        :param hyperpaths: Dic[origin: CityNode][destination: CityNode][StopNode] = List[List[ExtendedNodes]]
        :param Vij: dic[origin: CityNode][destination: CityNode] = vij
        :param assignment: dic[origin: CityNode][destination: CityNode][Stop: StopNode] = %V_OD
        :param successors: dic[origin: CityNode][destination: CityNode][ExtendedNode] = List[ExtendedEdge]
        :param extended_graph: ExtendedGraph object
        :param f: dict with frequency [veh/hr] for each route_id
        :param passenger_obj: Passenger obj
        :param z: boarding, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :param v: alighting, dic[route_id][direction][stop: StopNode] = pax [pax/veh]
        :return: float, users cost
        """
        ta, te, tv, t = self.resources_consumer(hyperpaths, Vij, assignment, successors,
                                                extended_graph, passenger_obj.va, f, z, v)
        pa = passenger_obj.pa
        pv = passenger_obj.pv
        pw = passenger_obj.pw
        pt = passenger_obj.pt

        return ta * pa + te * pw + tv * pv + t * pt



from __future__ import annotations

import operator
from typing import List

from sidermit.exceptions import CoIsNotValidExceptions, \
    C1IsNotValidExceptions, C2IsNotValidExceptions, VIsNotValidExceptions, \
    TIsNotValidExceptions, FmaxIsNotValidExceptions, \
    NameIsNotValidExceptions, KmaxIsNotValidExceptions, \
    TatIsNotValidExceptions, DIsNotValidExceptions, ByaIsNotValidExceptions, \
    ThetaIsNotValidExceptions, ModeDoesNotExistExceptions, \
    ModeNotFoundExceptions, AddModeExceptions, ModeIsNotValidException, FiniIsNotValidException


def mode_property(property_name, comp_function, exception_instance,
                  docstring=None):
    def getter(self):
        return self.__getattribute__(property_name)

    def setter(self, new_value):
        if new_value is None or comp_function(new_value, 0):
            raise exception_instance
        self.__setattr__(property_name, new_value)

    return property(getter, setter, doc=docstring)


class TransportMode:
    name = property(operator.attrgetter('_name'), doc="Name transport mode")
    bya = property(operator.attrgetter('_bya'),
                   doc="1 if passengers simultaneously get on and off the means of transport. 0 if they do it sequentially")
    co = mode_property('_co', operator.lt, CoIsNotValidExceptions(
        "You must give a value >=0 for co"), docstring="Unitary cost per vehicle per period of time [US$/h-veh]")
    c1 = mode_property('_c1', operator.lt, C1IsNotValidExceptions(
        "You must give a value >=0 for c1"), docstring="Unitary cost per seat per period of time [US$/h-veh]")
    c2 = mode_property('_c2', operator.lt, C2IsNotValidExceptions(
        "You must give a value >=0 for c2"), docstring="Unitary cost per vehicle per km of time [US$/h-veh]")
    v = mode_property('_v', operator.le, VIsNotValidExceptions(
        "You must give a value >=0 for v"), docstring="Cruise speed [km/h]")
    t = mode_property('_t', operator.lt, TIsNotValidExceptions(
        "You must give a value >=0 for t"), docstring="Boarding and alighting time [s/pax]")
    fmax = mode_property('_fmax', operator.lt, FmaxIsNotValidExceptions(
        "You must give a value >=0 for fmax"), docstring="Maximum frequency [veh/h]")
    kmax = mode_property('_kmax', operator.lt, KmaxIsNotValidExceptions(
        "You must give a vale >=0 for kmax"), docstring="Maximum vehicle size [pax/veh]")
    theta = property(operator.attrgetter('_theta'), doc="regularity of arrival of the mode of transport at the stops")
    tat = mode_property('_tat', operator.lt, TatIsNotValidExceptions(
        "You must give a value >=0 for tat"), docstring="Technological access time [min]")
    d = mode_property('_d', operator.lt, DIsNotValidExceptions(
        "You must give a value >=0 for d"), docstring="Parallel lines")
    fini = mode_property('_fini', operator.lt, FiniIsNotValidException(
        "You must give a value >=0 for d"), docstring="Initial frequency [veh/h]")

    @name.setter
    def name(self, value):
        if value is None:
            raise NameIsNotValidExceptions("You must give a name")
        self._name = value

    @bya.setter
    def bya(self, value):
        if value is None or value not in [0, 1]:
            raise ByaIsNotValidExceptions(
                "You must give a valid value for bya")
        self._bya = value

    @theta.setter
    def theta(self, value):
        if value is None or value < 0 or value > 1:
            raise ThetaIsNotValidExceptions(
                "You must give a value between [0-1] for theta")
        self._theta = value

    def __init__(self, name: str, bya: int, co: float, c1: float, c2: float,
                 v: float, t: float, fmax: float, kmax: float, theta: float,
                 tat: float, d: int, fini: float):
        self.name = name
        self.bya = bya
        self.co = co
        self.c1 = c1
        self.c2 = c2
        self.v = v
        self.t = t
        self.fmax = fmax
        self.kmax = kmax
        self.theta = theta
        self.tat = tat
        self.d = d
        self.fini = fini

    @staticmethod
    def get_default_modes() -> List[TransportMode]:
        """
        to get bus and metro mode of transport with default values
        :return: List[TransportMode]
        """
        bus_transport_mode = TransportMode("bus", 1, 8.61, 0.15, 0, 20, 2.5,
                                           150, 160, 0.7, 0, 4, 28)
        metro_transport_mode = TransportMode("metro", 0, 80.91, 0.3, 933.15,
                                             40, 0.33, 40, 1440, 0.5, 1, 1, 28)

        return [bus_transport_mode, metro_transport_mode]


class TransportModeManager:
    def __init__(self, add_default_mode: bool = True):
        """
        transport mode manager
        :param add_default_mode: (default: True) True to initialize list of modes with bus and metro with
         default values. False to initialize list of modes empty
        """

        self.__list_name = []
        self.__modes = []
        if add_default_mode:
            bus_obj, metro_obj = TransportMode.get_default_modes()
            self.__modes.append(bus_obj)
            self.__list_name.append(bus_obj.name)
            self.__modes.append(metro_obj)
            self.__list_name.append(metro_obj.name)

    def is_valid_to_assignment_step(self) -> bool:
        """
        to check that list of modes is valid to optimization. There should be only one or two transport mode.
        If there are two transport mode, one of them must have parameter d equal to 1
        :return: True if list of mode is valid to assignment step, False if not.
        """
        if len(self.__modes) == 1:
            return True

        if len(self.__modes) == 2:
            for mode in self.__modes:
                if mode.d == 1:
                    return True
        return False

    def get_mode(self, name: str) -> TransportMode:
        """
        to get a specific mode by name
        :param name: mode name
        :return: TransportMode
        """
        if name not in self.__list_name:
            raise ModeNotFoundExceptions("name mode not found")
        else:
            i = self.__list_name.index(name)
            mode = self.__modes[i]
            return mode

    def get_modes(self) -> List[TransportMode]:
        """
        to get all modes
        :return: List[TransportMode]
        """
        return self.__modes

    def get_modes_names(self) -> List[str]:
        """
        to get list_name of modes
        :return: List[names]
        """
        return self.__list_name

    def add_mode(self, mode_obj: TransportMode):
        """
        to add a new mode
        :param mode_obj: TransportMode
        :return:
        """
        if not isinstance(mode_obj, TransportMode):
            raise ModeIsNotValidException("mode_obj is not valid")

        if mode_obj.name in self.__list_name:
            raise AddModeExceptions("mode name exists, try with other name")

        self.__modes.append(mode_obj)
        self.__list_name.append(mode_obj.name)

    def remove_mode(self, name: str):
        """
        to delete a mode by name
        :param name: mode name
        :return:
        """

        if name in self.__list_name:
            i = self.__list_name.index(name)
            self.__list_name.pop(i)
            self.__modes.pop(i)
        else:
            raise ModeDoesNotExistExceptions("Mode does not exist")



import math
from collections import defaultdict
from enum import Enum
from typing import List

import networkx as nx
import pandas as pd
from matplotlib import pyplot as plt

from sidermit.city import Graph, CBD, Periphery, Subcenter
from sidermit.exceptions import *
from sidermit.publictransportsystem import TransportMode, TransportModeManager


class RouteType(Enum):
    """
    route types, CUSTOM for user-created custom routes, PREDEFINED for predefined routes created by a method of the
    TransportNetwork class, CIRCULAR to a special type of predefined route with different construction rules
    """
    CUSTOM = 1
    PREDEFINED = 2
    CIRCULAR = 3


class Route:

    def __init__(self, route_id, mode_obj: TransportMode, nodes_sequence_i: str, nodes_sequence_r: str,
                 stops_sequence_i: str, stops_sequence_r: str, _type: RouteType = RouteType.CUSTOM):
        """
        to defined a route
        :param route_id: route id
        :param mode_obj: TransportMode object. Means of transport of the route
        :param nodes_sequence_i: sequence of id nodes of the city graph where the route travels in the forward direction
        :param nodes_sequence_r: sequence of id nodes of the city graph where the route travels in the return direction
        :param stops_sequence_i: sequence of id nodes of the city graph where the route stops in the forward direction
        :param stops_sequence_r: sequence of id nodes of the city graph where the route stops in the return direction
        :param _type: RouteType, default value CUSTOM.
        """
        self.id = None
        self.mode = None
        self.nodes_sequence_i = None
        self.nodes_sequence_r = None
        self.stops_sequence_i = None
        self.stops_sequence_r = None
        self._type = _type
        if _type == RouteType.CUSTOM and (nodes_sequence_i is None or nodes_sequence_r is None):
            raise NodeSequencesIsNotValidException("You should give a value for nodes_sequence_i and nodes_sequence_r")
        if _type == RouteType.CUSTOM and (stops_sequence_i is None or stops_sequence_r is None):
            raise StopSequenceIsNotValidException("You should give a value for stops_sequence_i and stops_sequence_r")

        if route_id is None:
            raise RouteIdIsNotValidException("route_id is not valid. Try to give a value for route_id")

        # special case for circular routes
        if _type == RouteType.PREDEFINED or _type == RouteType.CIRCULAR:
            self.id = route_id
            self.mode = mode_obj
            self.nodes_sequence_i = self.sequences_to_list(nodes_sequence_i)
            self.nodes_sequence_r = self.sequences_to_list(nodes_sequence_r)
            self.stops_sequence_i = self.sequences_to_list(stops_sequence_i)
            self.stops_sequence_r = self.sequences_to_list(stops_sequence_r)

        else:
            # to valid parameters
            if self.parameters_validator(mode_obj,
                                         nodes_sequence_i, nodes_sequence_r,
                                         stops_sequence_i, stops_sequence_r):
                self.id = route_id
                self.mode = mode_obj
                self.nodes_sequence_i = self.sequences_to_list(nodes_sequence_i)
                self.nodes_sequence_r = self.sequences_to_list(nodes_sequence_r)
                self.stops_sequence_i = self.sequences_to_list(stops_sequence_i)
                self.stops_sequence_r = self.sequences_to_list(stops_sequence_r)

    def parameters_validator(self, mode_obj: TransportMode, nodes_sequence_i: str, nodes_sequence_r: str,
                             stops_sequence_i: str, stops_sequence_r: str) -> bool:
        """
        to check all parameters
        :param mode_obj: TransportMode object. Means of transport of the route
        :param nodes_sequence_i: sequence of id nodes of the city graph where the route travels in the forward direction
        :param nodes_sequence_r: sequence of id nodes of the city graph where the route travels in the return direction
        :param stops_sequence_i: sequence of id nodes of the city graph where the route stops in the forward direction
        :param stops_sequence_r: sequence of id nodes of the city graph where the route stops in the return direction
        :return: True if parameters are valid. Exception if not.
        """
        if not isinstance(mode_obj, TransportMode):
            raise ModeIsNotValidException("mode obj is not valid")

        nodes_i = self.sequences_to_list(nodes_sequence_i)
        nodes_r = self.sequences_to_list(nodes_sequence_r)
        stops_i = self.sequences_to_list(stops_sequence_i)
        stops_r = self.sequences_to_list(stops_sequence_r)

        if self.direction_validator(nodes_i, nodes_r):
            if self.stops_validator(nodes_i, stops_i) and self.stops_validator(nodes_r, stops_r):
                if self.sequences_validator(nodes_i) and self.sequences_validator(nodes_r):
                    return True

    @staticmethod
    def sequences_to_string(sequence_list: List[str]) -> str:
        """
        convert a node id sequence list to a string
        :param sequence_list: node id sequence list
        :return: String
        """
        line = ""
        for node in sequence_list:
            if line == "":
                line = line + str(node)
            else:
                line = line + "," + str(node)
        return line

    @staticmethod
    def sequences_to_list(sequence: str) -> List[int]:
        """
        convert a string of node id sequence to a list
        :param sequence: String
        :return: List[node id]
        """
        if sequence == "":
            return []

        nodes_split = sequence.split(",")
        nodes = []

        for node in nodes_split:
            nodes.append(int(node.rstrip("\n")))

        return nodes

    @staticmethod
    def stops_validator(nodes_list: List[int], stops_list: List[int]) -> bool:
        """
        to check if all stops of a direction of a route are a sub group of node_sequences. Also check if first and last
        nodes are stops
        :param nodes_list: list of node sequence
        :param stops_list: list of stops sequence
        :return: True if parameters are valid. Raise a exceptions if not
        """
        # to check if each stops be in node_sequences
        for stop in stops_list:
            if stop not in nodes_list:
                raise StopsSequencesException("stop is not reachable")
        # to check if first stop is equal to first node in node sequences
        if nodes_list[0] != stops_list[0]:
            raise FirstStopIsNotValidException("first stop is not valid, must be equal to first node")
        # to check if last stop is equal to last node in node sequences
        if nodes_list[len(nodes_list) - 1] != stops_list[len(stops_list) - 1]:
            raise LastStopIsNotValidException("last stop is not valid, must be equal to last node")
        return True

    @staticmethod
    def direction_validator(nodes_list_i: List[int], nodes_list_r: List[int]) -> bool:
        """
        to check if both direction of a route form a cycle
        :param nodes_list_i: list of node sequence (forward direction)
        :param nodes_list_r: list of node sequence (return direction)
        :return: True if parameters are valid. Raise a exceptions if not.
        """
        if nodes_list_i[0] != nodes_list_r[len(nodes_list_r) - 1] or \
                nodes_list_r[0] != nodes_list_i[len(nodes_list_i) - 1]:
            raise NotCycleException("sequence of nodes of both directions do not form a cycle")
        return True

    @staticmethod
    def sequences_validator(sequence: List[int]) -> bool:
        """
        to check if sequence have a loop
        :param sequence: list of node sequence
        :return: True if parameters are valid. Raise a exceptions if not.
        """

        for node1 in sequence:
            n = 0
            for node2 in sequence:
                if node1 == node2:
                    n = n + 1
            if n > 1:
                raise NodeSequencesIsNotValidException("node sequence loops")

        return True


class TransportNetwork:

    def __init__(self, graph_obj: Graph):
        """
        transport route manager on a city graph
        :param graph_obj: Graph where transport network develops
        """
        self.__graph_obj = graph_obj
        self.__routes = []
        self.__routes_id = []
        self.__modes = []

    def __edges_validator(self, node_list: List[int]) -> bool:
        """
        to check if each edges in a node_sequences list exist in the graph object
        :param node_list: list of nodes
        :return: True if parameters are valid. Raise a exceptions if not.
        """
        for i in range(len(node_list) - 1):
            j = i + 1
            if not self.__graph_obj.edge_exist(node_list[i], node_list[j]):
                raise NodeSequencesIsNotValidException("Node sequences is not valid because a edge does not exist")
        return True

    def get_modes(self) -> List[TransportMode]:
        return self.__modes

    def get_routes(self) -> List[Route]:
        """
        to get all routes
        :return: List[Route]
        """
        return self.__routes

    def get_route(self, route_id) -> Route:
        """
        to get a specific route by a route_id
        :param route_id: route id
        :return: Route object if route id is defined in the network. Raise a exceptions if not.
        """
        if route_id in self.__routes_id:
            i = self.__routes_id.index(route_id)
            return self.__routes[i]
        else:
            raise RouteIdNotFoundException("route_id not found")

    def add_transport_mode(self, mode: TransportMode):
        """
        to add a transport mode in the network
        :param mode: TransportMode
        :return:
        """

        if mode not in self.__modes:

            mode_manager = TransportModeManager(add_default_mode=False)

            list_mode = [mode]

            for modes in self.__modes:
                if modes not in list_mode:
                    list_mode.append(modes)

            for modes in list_mode:
                mode_manager.add_mode(modes)

            if mode_manager.is_valid_to_assignment_step():
                self.__modes.append(mode)
            else:
                raise TransportModeException(
                    "only 2 transport mode can be defined and at least one should have d parameter = 1")

    def remove_transport_mode(self, mode: TransportMode):
        """
        to remove a transport mode in the network and all lines defined with that transport mode
        :param mode: TransportMode
        :return:
        """

        list_mode = []

        for modes in self.__modes:
            if modes != mode:
                list_mode.append(modes)

        self.__modes = list_mode

        for route in self.__routes:
            if route.mode == mode:
                self.remove_route(route.id)

    def add_route(self, route_obj: Route):
        """
        to add a specific route in routes list
        :param route_obj: Route object
        :return:
        """
        if not isinstance(route_obj, Route):
            raise RouteIsNotvalidException("route_obj is not valid")

        route_id = route_obj.id
        nodes_sequence_i = route_obj.nodes_sequence_i
        nodes_sequence_r = route_obj.nodes_sequence_r

        if self.__edges_validator(nodes_sequence_i) or self.__edges_validator(nodes_sequence_r):
            if route_id not in self.__routes_id:
                self.add_transport_mode(route_obj.mode)
                self.__routes.append(route_obj)
                self.__routes_id.append(route_id)
            else:
                raise RouteIdDuplicatedException("route_id is duplicated")

    def remove_route(self, route_id):
        """
        to delete a specific route_id in the network
        :param route_id: route id
        :return:
        """
        if route_id in self.__routes_id:
            i = self.__routes_id.index(route_id)
            self.__routes_id.pop(i)
            self.__routes.pop(i)
        else:
            raise RouteIdNotFoundException("route_id not found")

    def routes_to_file(self, file_path):
        """
        to save file with all routes information
        :param file_path:
        :return:
        """
        # route_id, mode_name, nodes_sequence_i, nodes_sequence_r, stops_sequence_i, stops_sequence_r
        col_route_id = []
        col_mode = []
        col_nodes_sequence_i = []
        col_nodes_sequence_r = []
        col_stops_sequence_i = []
        col_stops_sequence_r = []

        for route in self.__routes:
            col_route_id.append(route.id)
            col_mode.append(route.mode.name)
            col_nodes_sequence_i.append(route.sequences_to_string(route.nodes_sequence_i))
            col_nodes_sequence_r.append(route.sequences_to_string(route.nodes_sequence_r))
            col_stops_sequence_i.append(route.sequences_to_string(route.stops_sequence_i))
            col_stops_sequence_r.append(route.sequences_to_string(route.stops_sequence_r))

        df_transit_network = pd.DataFrame()
        df_transit_network["route_id"] = col_route_id
        df_transit_network["mode"] = col_mode
        df_transit_network["nodes_sequence_i"] = col_nodes_sequence_i
        df_transit_network["nodes_sequence_r"] = col_nodes_sequence_r
        df_transit_network["stops_sequence_i"] = col_stops_sequence_i
        df_transit_network["stops_sequence_r"] = col_stops_sequence_r

        df_transit_network.to_csv(file_path, sep=";", index=False, encoding="utf-8")

    def get_circular_routes(self, mode_obj: TransportMode) -> List[Route]:
        """
        to get predefined circular routes, 2 routes with only a direction and whose stops and nodes sequence are all
        subcenter nodes.
        :param mode_obj: transport Mode
        :return: List[Route]
        """

        zones = self.__graph_obj.get_zones()
        if len(zones) <= 1:
            raise CircularRouteIsNotValidException("to add a predefined circular route you have a city "
                                                   "with more of one zone created")

        if not isinstance(mode_obj, TransportMode):
            raise ModeIsNotValidException("mode obj is not valid")

        mode_name = mode_obj.name

        route_id_i = "CIR_I_{}".format(mode_name)
        route_id_r = "CIR_R_{}".format(mode_name)

        nodes_sequence_i = ""
        nodes_sequence_r = ""

        for i in range(len(zones)):

            if nodes_sequence_i == "":
                nodes_sequence_i = nodes_sequence_i + str(zones[i].subcenter.id)
                nodes_sequence_r = nodes_sequence_r + str(zones[len(zones) - 1 - i].subcenter.id)
            else:
                nodes_sequence_i = nodes_sequence_i + "," + str(zones[i].subcenter.id)
                nodes_sequence_r = nodes_sequence_r + "," + str(zones[len(zones) - 1 - i].subcenter.id)

        nodes_sequence_i = nodes_sequence_i + "," + str(zones[0].subcenter.id)
        nodes_sequence_r = nodes_sequence_r + "," + str(zones[len(zones) - 1].subcenter.id)

        stops_sequence_i = nodes_sequence_i
        stops_sequence_r = nodes_sequence_r

        route1 = Route(route_id_i, mode_obj, nodes_sequence_i, "", stops_sequence_i, "", _type=RouteType.CIRCULAR)
        route2 = Route(route_id_r, mode_obj, "", nodes_sequence_r, "", stops_sequence_r, _type=RouteType.CIRCULAR)

        return [route1, route2]

    def get_feeder_routes(self, mode_obj: TransportMode) -> List[Route]:
        """
        to get predefined feeder routes, where for each zone exist a route with nodes and stops sequences beetween
        p-sc for I direction and sc-p for R direction.
        :param mode_obj: TransportMode
        :return: List[Route]
        """

        if not isinstance(mode_obj, TransportMode):
            raise ModeIsNotValidException("mode obj is not valid")

        mode_name = mode_obj.name

        routes = []
        for zone in self.__graph_obj.get_zones():
            id_p = zone.periphery.id
            id_sc = zone.subcenter.id

            route_id = "F_{}_{}".format(mode_name, zone.id)
            nodes_sequence_i = "{},{}".format(id_p, id_sc)
            stops_sequence_i = nodes_sequence_i
            nodes_sequence_r = "{},{}".format(id_sc, id_p)
            stops_sequence_r = nodes_sequence_r

            route = Route(route_id, mode_obj, nodes_sequence_i, nodes_sequence_r, stops_sequence_i,
                          stops_sequence_r, _type=RouteType.PREDEFINED)

            routes.append(route)

        return routes

    def get_radial_routes(self, mode_obj: TransportMode, short: bool = False, express: bool = False) -> List[Route]:
        """
        to get predefined radial routes, where for each zone exist a route with nodes and stops sequences beetween
        p-sc-cbd for I direction and cbd-sc-p for R direction.
        :param mode_obj: TransportMode
        :param short: if radial routes omit the passage through the periphery (default: False)
        :param express: if radial routes omit to stop in the subcenter (default: False)
        :return: List[Route]
        """

        if not isinstance(mode_obj, TransportMode):
            raise ModeIsNotValidException("mode obj is not valid")

        mode_name = mode_obj.name

        cbd = self.__graph_obj.get_cbd()
        id_cbd = cbd.id

        routes = []

        for zone in self.__graph_obj.get_zones():

            id_p = zone.periphery.id
            id_sc = zone.subcenter.id

            if short is True:
                route_id = "RS_{}_{}".format(mode_name, zone.id)
                nodes_sequence_i = "{},{}".format(id_sc, id_cbd)
                stops_sequence_i = nodes_sequence_i
                nodes_sequence_r = "{},{}".format(id_cbd, id_sc)
                stops_sequence_r = nodes_sequence_r
            else:
                if express is True:
                    route_id = "RE_{}_{}".format(mode_name, zone.id)
                    nodes_sequence_i = "{},{},{}".format(id_p, id_sc, id_cbd)
                    stops_sequence_i = "{},{}".format(id_p, id_cbd)
                    nodes_sequence_r = "{},{},{}".format(id_cbd, id_sc, id_p)
                    stops_sequence_r = "{},{}".format(id_cbd, id_p)
                else:
                    route_id = "R_{}_{}".format(mode_name, zone.id)
                    nodes_sequence_i = "{},{},{}".format(id_p, id_sc, id_cbd)
                    stops_sequence_i = nodes_sequence_i
                    nodes_sequence_r = "{},{},{}".format(id_cbd, id_sc, id_p)
                    stops_sequence_r = nodes_sequence_r

            route = Route(route_id, mode_obj, nodes_sequence_i, nodes_sequence_r, stops_sequence_i,
                          stops_sequence_r, _type=RouteType.PREDEFINED)
            routes.append(route)

        return routes

    def get_diametral_routes(self, mode_obj: TransportMode, jump: int = 1, short: bool = False,
                             express: bool = False) -> List[Route]:
        """
        to get predefined diametral routes, where for each zone exist a route with nodes and stops sequences beetween
        p-sc-cbd-sc'-p' for I direction and p'-sc'-cbd-sc-p for R direction. p' and sc' are periphery and subcenter
        nodes with zone id equivalent to the id of the treated zone plus the jump
        :param mode_obj: TransportMode
        :param jump: to identified other zone where diametral routes transit
        :param short: if diametral routes omit the passage through the periphery (default: False)
        :param express: if diametral routes omit to stop in the subcenter and cbd nodes  (default: False)
        :return: List[Route]
        """
        if not isinstance(mode_obj, TransportMode):
            raise ModeIsNotValidException("mode obj is not valid")

        mode_name = mode_obj.name

        cbd = self.__graph_obj.get_cbd()
        id_cbd = cbd.id

        zones = self.__graph_obj.get_zones()

        if jump > len(zones) / 2 or jump <= 0 or not isinstance(jump, int):
            raise JumpIsNotValidException("jump must be a int in range (0-n° zones)")

        end = len(zones)
        if len(zones) % 2.0 == 0 and jump == math.floor(len(zones) / 2):
            end = math.floor(len(zones) / 2)

        routes = []

        for zone in zones[:end]:
            id_p = zone.periphery.id
            id_sc = zone.subcenter.id

            zone_id = zone.id

            if zone_id + jump <= len(zones):
                zone2_id = zone_id + jump
            else:
                zone2_id = zone_id + jump - len(zones)

            # zones are sort per id in list
            zone2 = zones[zone2_id - 1]

            id_p2 = zone2.periphery.id
            id_sc2 = zone2.subcenter.id

            if short is True:
                if express is True:
                    route_id = "DSE{}_{}_{}".format(jump, mode_name, zone.id)
                    nodes_sequence_i = "{},{},{}".format(id_sc, id_cbd, id_sc2)
                    stops_sequence_i = "{},{}".format(id_sc, id_sc2)
                    nodes_sequence_r = "{},{},{}".format(id_sc2, id_cbd, id_sc)
                    stops_sequence_r = "{},{}".format(id_sc2, id_sc)
                else:
                    route_id = "DS{}_{}_{}".format(jump, mode_name, zone.id)
                    nodes_sequence_i = "{},{},{}".format(id_sc, id_cbd, id_sc2)
                    stops_sequence_i = nodes_sequence_i
                    nodes_sequence_r = "{},{},{}".format(id_sc2, id_cbd, id_sc)
                    stops_sequence_r = nodes_sequence_r
            else:
                if express is True:
                    route_id = "DE{}_{}_{}".format(jump, mode_name, zone.id)
                    nodes_sequence_i = "{},{},{},{},{}".format(id_p, id_sc, id_cbd, id_sc2, id_p2)
                    stops_sequence_i = "{},{}".format(id_p, id_p2)
                    nodes_sequence_r = "{},{},{},{},{}".format(id_p2, id_sc2, id_cbd, id_sc, id_p)
                    stops_sequence_r = "{},{}".format(id_p2, id_p)
                else:
                    route_id = "D{}_{}_{}".format(jump, mode_name, zone.id)
                    nodes_sequence_i = "{},{},{},{},{}".format(id_p, id_sc, id_cbd, id_sc2, id_p2)
                    stops_sequence_i = nodes_sequence_i
                    nodes_sequence_r = "{},{},{},{},{}".format(id_p2, id_sc2, id_cbd, id_sc, id_p)
                    stops_sequence_r = nodes_sequence_r
            route = Route(route_id, mode_obj, nodes_sequence_i, nodes_sequence_r, stops_sequence_i,
                          stops_sequence_r, _type=RouteType.PREDEFINED)
            routes.append(route)

        return routes

    def get_tangencial_routes(self, mode_obj: TransportMode, jump: int = 1, short: bool = False,
                              express: bool = False) -> List[Route]:
        """
        to get predefined tangencial routes, where for each zone exist a route with nodes and stops sequences beetween
        p-sc-sc'-...-sc''-p'' for I direction and p''-sc''-...-sc'-sc-p for R direction. sc', ..., sc'' are subcenter
        nodes with zone id less than or equal to id of the treated zone plus the jump. p'' and sc'' are periphery
        and subcenter nodes with id equivalent to the id of the treated zone plus the jump
        :param mode_obj: TransportMode
        :param jump: to identified other zone where tangencial routes transit
        :param short: if radial routes omit the passage through the periphery (default: False)
        :param express: if radial routes omit to stop in the subcenters nodes (default: False)
        :return: List[Route]
        """
        if not isinstance(mode_obj, TransportMode):
            raise ModeIsNotValidException("mode obj is not valid")

        mode_name = mode_obj.name

        zones = self.__graph_obj.get_zones()

        if jump > len(zones) / 2 or jump <= 0 or not isinstance(jump, int):
            raise JumpIsNotValidException("jump must be a int in range (0-n° zones)")

        end = len(zones)
        if len(zones) % 2.0 == 0 and jump == math.floor(len(zones) / 2):
            end = math.floor(len(zones) / 2)

        routes = []

        for zone in zones[:end]:
            id_p = zone.periphery.id
            id_sc = zone.subcenter.id

            zone_id = zone.id

            if zone_id + jump <= len(zones):
                zone2_id = zone_id + jump
            else:
                zone2_id = zone_id + jump - len(zones)

            # zones are sort per id in list
            zone2 = zones[zone2_id - 1]

            id_p2 = zone2.periphery.id
            id_sc2 = zone2.subcenter.id

            list_zones_id = []

            if zone2_id > zone_id and jump != 1:
                for i in range(zone_id + 1, zone2_id):
                    list_zones_id.append(i)

            if zone2_id < zone_id:
                if zone_id != len(zones):
                    for i in range(zone_id + 1, len(zones) + 1):
                        list_zones_id.append(i)

                for i in range(1, zone2_id):
                    list_zones_id.append(i)

            if short is True:
                if express is True:
                    route_id = "TSE{}_{}_{}".format(jump, mode_name, zone.id)

                    nodes_sequence_i = "{}".format(id_sc)
                    nodes_sequence_r = "{}".format(id_sc2)

                    for i in range(len(list_zones_id)):
                        zone3 = zones[list_zones_id[i] - 1]
                        zone4 = zones[list_zones_id[len(list_zones_id) - 1 - i] - 1]

                        nodes_sequence_i = nodes_sequence_i + "," + str(zone3.subcenter.id)
                        nodes_sequence_r = nodes_sequence_r + "," + str(zone4.subcenter.id)

                    nodes_sequence_i = nodes_sequence_i + "," + str("{}".format(id_sc2))
                    nodes_sequence_r = nodes_sequence_r + "," + str("{}".format(id_sc))

                    stops_sequence_i = "{},{}".format(id_sc, id_sc2)
                    stops_sequence_r = "{},{}".format(id_sc2, id_sc)

                else:
                    route_id = "TS{}_{}_{}".format(jump, mode_name, zone.id)
                    nodes_sequence_i = "{}".format(id_sc)
                    nodes_sequence_r = "{}".format(id_sc2)
                    for i in range(len(list_zones_id)):
                        zone3 = zones[list_zones_id[i] - 1]
                        zone4 = zones[list_zones_id[len(list_zones_id) - 1 - i] - 1]

                        nodes_sequence_i = nodes_sequence_i + "," + str(zone3.subcenter.id)
                        nodes_sequence_r = nodes_sequence_r + "," + str(zone4.subcenter.id)
                    nodes_sequence_i = nodes_sequence_i + "," + "{}".format(id_sc2)
                    nodes_sequence_r = nodes_sequence_r + "," + "{}".format(id_sc)
                    stops_sequence_i = nodes_sequence_i
                    stops_sequence_r = nodes_sequence_r

            else:
                if express is True:
                    route_id = "TE{}_{}_{}".format(jump, mode_name, zone.id)
                    nodes_sequence_i = "{},{}".format(id_p, id_sc)
                    nodes_sequence_r = "{},{}".format(id_p2, id_sc2)
                    for i in range(len(list_zones_id)):
                        zone3 = zones[list_zones_id[i] - 1]
                        zone4 = zones[list_zones_id[len(list_zones_id) - 1 - i] - 1]

                        nodes_sequence_i = nodes_sequence_i + "," + str(zone3.subcenter.id)
                        nodes_sequence_r = nodes_sequence_r + "," + str(zone4.subcenter.id)
                    nodes_sequence_i = nodes_sequence_i + "," + "{},{}".format(id_sc2, id_p2)
                    nodes_sequence_r = nodes_sequence_r + "," + "{},{}".format(id_sc, id_p)
                    stops_sequence_i = "{},{}".format(id_p, id_p2)
                    stops_sequence_r = "{},{}".format(id_p2, id_p)
                else:
                    route_id = "T{}_{}_{}".format(jump, mode_name, zone.id)
                    nodes_sequence_i = "{},{}".format(id_p, id_sc)
                    nodes_sequence_r = "{},{}".format(id_p2, id_sc2)
                    for i in range(len(list_zones_id)):
                        zone3 = zones[list_zones_id[i] - 1]
                        zone4 = zones[list_zones_id[len(list_zones_id) - 1 - i] - 1]

                        nodes_sequence_i = nodes_sequence_i + "," + str(zone3.subcenter.id)
                        nodes_sequence_r = nodes_sequence_r + "," + str(zone4.subcenter.id)
                    nodes_sequence_i = nodes_sequence_i + "," + "{},{}".format(id_sc2, id_p2)
                    nodes_sequence_r = nodes_sequence_r + "," + "{},{}".format(id_sc, id_p)
                    stops_sequence_i = nodes_sequence_i
                    stops_sequence_r = nodes_sequence_r

            route = Route(route_id, mode_obj, nodes_sequence_i, nodes_sequence_r, stops_sequence_i,
                          stops_sequence_r, _type=RouteType.PREDEFINED)
            routes.append(route)

        return routes

    def plot(self, file_path, list_routes=None, direction=None):
        """
        to plot network and graph
        :param file_path:
        :param list_routes: list of routes to plot, default value is None and plot all routes
        :param direction: "I" or "R" direction to plot. default value is None and plot both direction.
        :return:
        """
        # if list routes is empty then plot all routes
        if list_routes is None:
            list_routes = self.__routes_id

        # edges information and positions
        # edge city
        edges_graph = []
        position = defaultdict(list)
        for edge in self.__graph_obj.get_edges():
            edges_graph.append((edge.node1.id, edge.node2.id))
            if not position.get(edge.node1.id):
                position[edge.node1.id].append(edge.node1.x)
                position[edge.node1.id].append(edge.node1.y)
            if not position.get(edge.node2.id):
                position[edge.node2.id].append(edge.node2.x)
                position[edge.node2.id].append(edge.node2.y)

        # nodes information and positions
        info_cbd = []
        info_sc = []
        info_p = []
        id_cbd = []
        id_sc = []
        id_p = []
        x_cbd = []
        y_cbd = []
        x_sc = []
        y_sc = []
        x_p = []
        y_p = []

        for node in self.__graph_obj.get_nodes():
            if isinstance(node, CBD):
                info_cbd.append(node)
            if isinstance(node, Periphery):
                info_p.append(node)
            if isinstance(node, Subcenter):
                info_sc.append(node)

        for cbd in info_cbd:
            x_cbd.append(cbd.x)
            y_cbd.append(cbd.y)
            id_cbd.append(cbd.id)

        for sc in info_sc:
            x_sc.append(sc.x)
            y_sc.append(sc.y)
            id_sc.append(sc.id)
        for p in info_p:
            x_p.append(p.x)
            y_p.append(p.y)
            id_p.append(p.id)

        # edges routes and stops
        edges_i = []
        edges_r = []
        stops_i = []
        stops_r = []
        for route_id in list_routes:
            if route_id not in self.__routes_id:
                raise RouteIdNotFoundException("route_id does not found")
            else:
                ind = self.__routes_id.index(route_id)
                route = self.__routes[ind]
                nodes_i = route.nodes_sequence_i
                nodes_r = route.nodes_sequence_r
                stop_i = route.stops_sequence_i
                stop_r = route.stops_sequence_r

                for i in range(len(nodes_i) - 1):
                    id1 = nodes_i[i]
                    id2 = nodes_i[i + 1]
                    edges_i.append((id1, id2))
                for i in range(len(nodes_r) - 1):
                    id1 = nodes_r[i]
                    id2 = nodes_r[i + 1]
                    edges_r.append((id1, id2))
                for i in range(len(stop_i)):
                    if stop_i[i] not in stops_i:
                        stops_i.append(stop_i[i])
                for i in range(len(stop_r)):
                    if stop_r[i] not in stops_r:
                        stops_r.append(stop_r[i])

        G = nx.DiGraph()
        G.add_edges_from(edges_graph)
        G.add_edges_from(edges_i)
        G.add_edges_from(edges_r)

        # separate calls to draw labels, nodes and edges
        # plot p, Sc and CBD
        nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=id_p, node_color='red', node_size=300)
        nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=id_sc, node_color='blue', node_size=300)
        nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=id_cbd, node_color='purple',
                               node_size=300)
        # plot stops
        if direction is None or direction == "I":
            nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=stops_i, node_color='yellow',
                                   node_size=300)
        if direction is None or direction == "R":
            nx.draw_networkx_nodes(G, position, cmap=plt.get_cmap('Set2'), nodelist=stops_r, node_color='yellow',
                                   node_size=300)
        # plot labels
        nx.draw_networkx_labels(G, position)
        # plot edges city
        nx.draw_networkx_edges(G, position, edgelist=edges_graph, edge_color='orange', arrows=True)
        # plot edges_i
        if direction is None or direction == "I":
            nx.draw_networkx_edges(G, position, edgelist=edges_i, edge_color='lime', arrows=True)
        # plot edges_r
        if direction is None or direction == "R":
            nx.draw_networkx_edges(G, position, edgelist=edges_r, edge_color='aqua', arrows=True)

        plt.title("City graph")
        plt.xlabel("X")
        plt.ylabel("Y")
        plt.gca().set_aspect('equal')
        plt.savefig(file_path)



from .operators_cost import OperatorsCost
from .infrastructure_cost import InfrastructureCost
from .constrains import Constrains
from .users_cost import UsersCost
from .optimizer import Optimizer

__all__ = ['OperatorsCost', 'Constrains', 'InfrastructureCost', 'UsersCost', 'Optimizer']


import filecmp
import os
import unittest
from pathlib import Path

from sidermit import exceptions
from sidermit.city import graph


class test_graph(unittest.TestCase):

    def setUp(self):
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        self.data_path = os.path.join(self.dir_path, 'file')
        self.data_path = os.path.join(self.data_path, 'graph')
        self.test_0zone_path = os.path.join(self.data_path, 'test_0zone.PAJEK')
        self.test_1zone_path = os.path.join(self.data_path, 'test_1zone.PAJEK')
        self.test_2zone_path = os.path.join(self.data_path, 'test_2zone.PAJEK')
        self.test_3zone_path = os.path.join(self.data_path, 'test_3zone.PAJEK')
        self.test_formatPajekExceptions = os.path.join(self.data_path,
                                                       'test_formatPajekExceptions.PAJEK')
        self.test_NodeTypeExceptions = os.path.join(self.data_path,
                                                    'test_NodeTypeExceptions.PAJEK')
        self.test_NumberOfLinesExceptions = os.path.join(self.data_path,
                                                         'test_NumberOfLinesExceptions.PAJEK')
        self.test_NumberNodesForZoneExceptions = os.path.join(self.data_path,
                                                              'test_NumberNodesForZoneExceptions.PAJEK')
        self.test_needCbdExceptions = os.path.join(self.data_path,
                                                   'test_needCbdExceptions.PAJEK')
        self.test_nodeIdExceptions = os.path.join(self.data_path,
                                                  'test_nodeIdExceptions.PAJEK')

        self.test_nodeIdStringExceptions = os.path.join(self.data_path,
                                                        'test_node_id_string_exceptions.PAJEK')
        self.test_ZoneIdStringExceptions = os.path.join(self.data_path,
                                                        'test_zone_id_string_exceptions.PAJEK')

    def test_exceptions_string_in_pajek(self):
        """
        test exceptions string to node id and zone id in pajek
        :return:
        """
        with self.assertRaises(exceptions.NodeIdIsNotAnInteger):
            graph.Graph.build_from_file(self.test_nodeIdStringExceptions)
        with self.assertRaises(exceptions.ZoneIdIsNotAnInteger):
            graph.Graph.build_from_file(self.test_ZoneIdStringExceptions)

    def test_get_method(self):
        """
        test methods get class graph
        :return:
        """
        g = graph.Graph.build_from_file(self.test_3zone_path)

        self.assertTrue(isinstance(g.get_cbd(), graph.CBD))
        self.assertEqual(g.get_n(), 3)
        self.assertEqual(len(g.get_zones()), 3)
        self.assertEqual(len(g.get_nodes()), 7)
        self.assertEqual(len(g.get_edges()), 18)

    def test_get_edges_distance(self):
        """
        to test get_edges_distance method of class Graph
        :return:
        """
        graph_obj = graph.Graph.build_from_parameters(2, 10, 0.5, 2)

        edge_distance = graph_obj.get_edges_distance()

        self.assertEqual(edge_distance[1][2], 5)
        self.assertEqual(edge_distance[2][0], 10)

    def test_Node_Exceptions(self):
        """
        to test exceptions class Node
        :return:
        """
        node_id = 0
        x = 0
        y = 0
        radius = 0
        angle = 0
        width = 0
        zone_id = 0
        name = "CBD"
        with self.assertRaises(exceptions.NameIsNotDefinedException):
            graph.Node(node_id, x, y, radius, angle, width, zone_id, None)
        with self.assertRaises(exceptions.NodeIdIsNotValidException):
            graph.Node(None, x, y, radius, angle, width, zone_id, name)
        with self.assertRaises(exceptions.ZoneIdIsNotValidException):
            graph.Node(node_id, x, y, radius, angle, width, None, name)
        with self.assertRaises(exceptions.NodeRadiusIsNotValidException):
            graph.Node(node_id, x, y, -2, angle, width, zone_id, name)
        with self.assertRaises(exceptions.NodeAngleIsNotValidException):
            graph.Node(node_id, x, y, radius, -10, width, zone_id, name)
        with self.assertRaises(exceptions.NodeWidthIsNotValidException):
            graph.Node(node_id, x, y, radius, angle, -2, zone_id, name)

    def test_CBD_Exceptions(self):
        """
        to test exceptions class CBD
        :return:
        """
        node_id = 0
        x = 0
        y = 0
        radius = 0
        angle = 0
        width = 0
        name = "CBD"
        with self.assertRaises(exceptions.ZoneIdIsNotValidException):
            graph.CBD(node_id, x, y, radius, angle, width, 2, name)

    def test_Zone_Exceptions(self):
        """
        to test exceptions class zone
        :return:
        """
        with self.assertRaises(exceptions.ZoneIdIsNotValidException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            sc = graph.Subcenter(2, 1, 1, 1, 1, 1, 1, "sc")
            graph.Zone(-1, p, sc)
        with self.assertRaises(exceptions.ZoneIdIsNotValidException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            sc = graph.Subcenter(2, 1, 1, 1, 1, 1, 1, "sc")
            graph.Zone(None, p, sc)
        with self.assertRaises(
                exceptions.PeripheryNodeTypeIsNotValidException):
            sc = graph.Subcenter(2, 1, 1, 1, 1, 1, 1, "sc")
            graph.Zone(1, sc, sc)
        with self.assertRaises(
                exceptions.SubcenterNodeTypeIsNotValidException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            graph.Zone(1, p, p)
        with self.assertRaises(exceptions.ZoneIdIsNotValidException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            sc = graph.Subcenter(2, 1, 1, 1, 1, 1, 1, "sc")
            graph.Zone(2, p, sc)

    def test_Edge_Exceptions(self):
        """
        to test exceptions class Edge
        :return:
        """
        with self.assertRaises(exceptions.EdgeIdIsNotValidException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            sc = graph.Subcenter(2, 1, 1, 1, 1, 1, 1, "sc")
            graph.Edge(None, p, sc)
        with self.assertRaises(exceptions.EdgeIsNotAvailableException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            graph.Edge(1, p, p)
        with self.assertRaises(exceptions.EdgeIsNotAvailableException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            cbd = graph.CBD(0, 0, 0, 0, 0, 0, 0, "CBD")
            graph.Edge(1, p, cbd)
        with self.assertRaises(exceptions.EdgeIsNotAvailableException):
            p = graph.Periphery(1, 1, 1, 1, 1, 1, 1, "p")
            cbd = graph.CBD(0, 0, 0, 0, 0, 0, 0, "CBD")
            graph.Edge(1, cbd, p)

    def test_build_from_pajek_file(self):
        """
        to test reading pajek file
        :return:
        """
        g = graph.Graph.build_from_file(self.test_0zone_path)

        self.assertEqual(len(g.get_nodes()), 1)
        self.assertEqual(len(g.get_edges()), 0)
        self.assertEqual(len(g.get_zones()), 0)

        g = graph.Graph.build_from_file(self.test_1zone_path)

        self.assertEqual(len(g.get_nodes()), 3)
        self.assertEqual(len(g.get_edges()), 4)
        self.assertEqual(len(g.get_zones()), 1)

        g = graph.Graph.build_from_file(self.test_2zone_path)

        self.assertEqual(len(g.get_nodes()), 5)
        self.assertEqual(len(g.get_edges()), 10)
        self.assertEqual(len(g.get_zones()), 2)

        g = graph.Graph.build_from_file(self.test_3zone_path)

        self.assertEqual(len(g.get_nodes()), 7)
        self.assertEqual(len(g.get_edges()), 18)
        self.assertEqual(len(g.get_zones()), 3)

    def test_build_from_parameters(self):
        """
        to test construction from parameters
        :return:
        """
        g = graph.Graph.build_from_parameters(1, 1000, 0.5, 0)

        self.assertEqual(len(g.get_nodes()), 3)
        self.assertEqual(len(g.get_edges()), 4)
        self.assertEqual(len(g.get_zones()), 1)

        n, L, g, P, etha, etha_zone, angles, Gi, Hi = g.get_parameters()

        self.assertEqual(n, 1)
        self.assertEqual(L, 1000)
        self.assertEqual(g, 0.5)
        self.assertEqual(P, 0)

        g = graph.Graph.build_from_parameters(1, 1000, 0.5, 0, angles=[10])

        self.assertEqual(len(g.get_nodes()), 3)
        self.assertEqual(len(g.get_edges()), 4)
        self.assertEqual(len(g.get_zones()), 1)

        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 0, angles=[10, 50])

        self.assertEqual(len(g.get_nodes()), 5)
        self.assertEqual(len(g.get_edges()), 10)
        self.assertEqual(len(g.get_zones()), 2)

        g = graph.Graph.build_from_parameters(1, 1000, 0.5, 0, Gi=[2])

        self.assertEqual(len(g.get_nodes()), 3)
        self.assertEqual(len(g.get_edges()), 4)
        self.assertEqual(len(g.get_zones()), 1)

        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 0, Gi=[2, 0.5])

        self.assertEqual(len(g.get_nodes()), 5)
        self.assertEqual(len(g.get_edges()), 10)
        self.assertEqual(len(g.get_zones()), 2)

        g = graph.Graph.build_from_parameters(1, 1000, 0.5, 0, Hi=[2])

        self.assertEqual(len(g.get_nodes()), 3)
        self.assertEqual(len(g.get_edges()), 4)
        self.assertEqual(len(g.get_zones()), 1)

        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 0, Hi=[2, 0.5])

        self.assertEqual(len(g.get_nodes()), 5)
        self.assertEqual(len(g.get_edges()), 10)
        self.assertEqual(len(g.get_zones()), 2)

        g = graph.Graph.build_from_parameters(1, 1000, 0.5, 0, etha=0.5,
                                              etha_zone=1)

        self.assertEqual(len(g.get_nodes()), 3)
        self.assertEqual(len(g.get_edges()), 4)
        self.assertEqual(len(g.get_zones()), 1)

        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 0, etha=0.5,
                                              etha_zone=2)

        self.assertEqual(len(g.get_nodes()), 5)
        self.assertEqual(len(g.get_edges()), 10)
        self.assertEqual(len(g.get_zones()), 2)

        g = graph.Graph.build_from_parameters(1, 1000, 0.5, 0, etha=0.5,
                                              etha_zone=1, angles=[10], Gi=[2],
                                              Hi=[2])

        self.assertEqual(len(g.get_nodes()), 3)
        self.assertEqual(len(g.get_edges()), 4)
        self.assertEqual(len(g.get_zones()), 1)

        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 0, etha=0.5,
                                              etha_zone=1, angles=[10, 50],
                                              Gi=[2, 0.5],
                                              Hi=[2, 0.5])

        self.assertEqual(len(g.get_nodes()), 5)
        self.assertEqual(len(g.get_edges()), 10)
        self.assertEqual(len(g.get_zones()), 2)

        n, L, g, P, etha, etha_zone, angles, Gi, Hi = g.get_parameters()

        self.assertEqual(n, 2)
        self.assertEqual(L, 1000)
        self.assertEqual(g, 0.5)
        self.assertEqual(P, 0)
        self.assertEqual(etha, 0.5)
        self.assertEqual(etha_zone, 1)
        self.assertEqual(angles, [10, 50])
        self.assertEqual(Gi, [2, 0.5])
        self.assertEqual(Hi, [2, 0.5])

    def test_raises_fileformat(self):
        """
        to test pajek file exceptions, should have 7 columns separated by
        white space
        :return:
        """
        with self.assertRaises(exceptions.FileFormatIsNotValidException):
            graph.Graph.build_from_file(self.test_2zone_path,
                                        file_format="CSV")

    def test_raises_payekformat(self):
        """
        to test pajek format exceptions: file must have 7 columns, separator
        for " "
        :return:
        """
        with self.assertRaises(exceptions.PajekFormatIsNotValidException):
            graph.Graph.build_from_file(self.test_formatPajekExceptions)

    def test_symmetric_parameters_exceptions(self):
        """
        to test symmetric parameters exceptions n, l, g, p
        :return:
        """
        with self.assertRaises(exceptions.NIsNotValidNumberException):
            graph.Graph.build_from_parameters(-1, 1000, 0.5, 0)
        with self.assertRaises(exceptions.LIsNotValidNumberException):
            graph.Graph.build_from_parameters(1, -1000, 0.5, 0)
        with self.assertRaises(exceptions.GIsNotValidNumberException):
            graph.Graph.build_from_parameters(1, 1000, -0.5, 0)
        with self.assertRaises(exceptions.PIsNotValidNumberException):
            graph.Graph.build_from_parameters(1, 1000, 0.5, -1)

    def test_asymmetric_parameters_exceptions(self):
        """
        to test asymmetric parameters exceptions etha, etha_zone, angles, Gi,
        Hi
        :return:
        """
        with self.assertRaises(exceptions.EthaValueRequiredException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, etha_zone=2)
        with self.assertRaises(exceptions.EthaZoneValueRequiredException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, etha=0.5)
        with self.assertRaises(exceptions.EthaValueIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, etha=-1,
                                              etha_zone=2)
        with self.assertRaises(exceptions.EthaZoneValueIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, etha=0.5,
                                              etha_zone=3)
        with self.assertRaises(exceptions.AngleListLengthIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, angles=[20])
        with self.assertRaises(exceptions.AngleValueIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0,
                                              angles=[20, 370])
        with self.assertRaises(exceptions.GiListLengthIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, Gi=[0.5])
        with self.assertRaises(exceptions.GiValueIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, Gi=[0.5, -2])
        with self.assertRaises(exceptions.HiListLengthIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, Hi=[0.5])
        with self.assertRaises(exceptions.HiValueIsNotValidException):
            graph.Graph.build_from_parameters(2, 1000, 0.5, 0, Hi=[0.5, -2])

    def test_graph_to_pajek(self):
        """
        to test file writing
        :return:
        """
        filename = 'write1.PAJEK'
        # build graph
        g = graph.Graph.build_from_parameters(2, 1000, 0.5, 0, etha=0.5,
                                              etha_zone=1, angles=[10, 50],
                                              Gi=[2, 0.5],
                                              Hi=[2, 0.5])
        # write file
        g.graph_to_pajek_file(os.path.join(self.data_path, filename), graph.GraphContentFormat.PAJEK)
        file_obj = Path(os.path.join(self.data_path, filename))

        # test build
        self.assertTrue(file_obj.is_file())

        # to compare file with a test file
        self.assertTrue(
            filecmp.cmp(os.path.join(self.data_path, filename),
                        os.path.join(self.data_path, "write1_test.PAJEK")))
        # remove file
        os.remove(os.path.join(self.data_path, filename))

    def test_obtain_angle(self):
        """
        to test assistant method Graph.obtain_angle(x, y)
        :return:
        """
        g = graph.Graph()
        self.assertEqual(int(g.get_angle(1, 0)), 0)
        self.assertEqual(int(g.get_angle(0, 1)), 90)
        self.assertEqual(int(g.get_angle(-1, 0)), 180)
        self.assertEqual(int(g.get_angle(0, -1)), 270)
        self.assertEqual(int(g.get_angle(1, 1)), 45)
        self.assertEqual(int(g.get_angle(-1, 1)), 135)
        self.assertEqual(int(g.get_angle(-1, -1)), 225)
        self.assertEqual(int(g.get_angle(1, -1)), 315)

    def test_pajekfile_to_dataframe_exceptions(self):
        """
        to test exception of node type in pajek file
        :return:
        """
        with self.assertRaises(exceptions.NodeTypeIsNotValidException):
            graph.Graph.build_from_file(self.test_NodeTypeExceptions)

    def test_build_from_file_exceptions(self):
        """
        to test exception of number of lines in pajek file
        :return:
        """
        with self.assertRaises(exceptions.LineNumberInFileIsNotValidException):
            graph.Graph.build_from_file(self.test_NumberOfLinesExceptions)

        with self.assertRaises(
                exceptions.PeripherySubcenterNumberForZoneException):
            graph.Graph.build_from_file(self.test_NumberNodesForZoneExceptions)

    def test_add_zone_exceptions(self):
        """
        to check exceptions __add_zones method
        :return:
        """
        with self.assertRaises(exceptions.CBDDoesNotExistException):
            graph.Graph.build_from_file(self.test_needCbdExceptions)

    def test_add_nodes_exceptions(self):
        """
        to check exceptions __add_nodes method
        :return:
        """
        with self.assertRaises(exceptions.NodeIdDuplicatedException):
            graph.Graph.build_from_file(self.test_nodeIdExceptions)

    def test_edges_exist(self):
        """
        to test edge_exist method
        :return:
        """
        g = graph.Graph.build_from_parameters(7, 1000, 0.5, 0,
                                              angles=[10, 50, 150, 180, 270,
                                                      300, 320], etha=0.5,
                                              etha_zone=3,
                                              Hi=[1, 2, 1, 1, 1, 0.5, 3],
                                              Gi=[1, 2, 1, 1, 1, 3, 2])
        self.assertTrue(g.edge_exist(1, 2))
        self.assertTrue(not g.edge_exist("1", "200"))

    # def test_plot(self):
    #     """
    #     to test plot method
    #     :return:
    #     """
    #     g = graph.Graph.build_from_parameters(7, 1000, 0.5, 0,
    #                                           angles=[10, 50, 150, 180, 270,
    #                                                   300, 320], etha=0.5,
    #                                           etha_zone=3,
    #                                           Hi=[1, 2, 1, 1, 1, 0.5, 3],
    #                                           Gi=[1, 2, 1, 1, 1, 3, 2])
    #
    #     # save figure in path
    #     g.plot(os.path.join(self.data_path, 'figure1_test.png'))
    #     fileObj = Path(os.path.join(self.data_path, 'figure1_test.png'))
    #     # test
    #     self.assertTrue(fileObj.is_file())
    #     # to compare figure with a test figure
    #     self.assertTrue(
    #         filecmp.cmp(os.path.join(self.data_path, 'figure_test.png'),
    #                     os.path.join(self.data_path, "figure1_test.png")))
    #
    #     # remove file
    #     os.remove(os.path.join(self.data_path, 'figure1_test.png'))
